/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 2003 - 2019     SEGGER Microcontroller GmbH              *
*                                                                    *
*       www.segger.com     Support: www.segger.com/ticket            *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host * USB Host stack for embedded applications        *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product.                          *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host version: V2.16                                    *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------

File    : USBH_BULK_ListDevices.c
Purpose : Shows emUSB-Host Bulk API usage.

Additional information:
  Preparations:
    None.

  Expected behavior:
    When a device is connected the application print some information
    about that device.

  Sample output:
    11:007 USBH_Task - **** New Device connected [0]
    11:007 USBH_Task -    Vendor  Id = 0x8765
    11:007 USBH_Task -    Product Id = 0x1089
    11:007 USBH_Task -    Interface  = 0
    11:007 USBH_Task -    Serial no. = 13245678
    11:007 USBH_Task -    Speed      = HighSpeed
    11:007 USBH_Task -    Endpoints:
    11:007 USBH_Task -       Addr 81, Type = BULK
    11:007 USBH_Task -       Addr 01, Type = BULK
*/

/*********************************************************************
*
*       #include section
*
**********************************************************************
*/
#include <stdio.h>
#include "RTOS.h"
#include "BSP.h"
#include "USBH.h"
#include "USBH_BULK.h"

/*********************************************************************
*
*       Local defines, configurable
*
**********************************************************************
*/

enum {
  TASK_PRIO_APP = 150,
  TASK_PRIO_USBH_MAIN,
  TASK_PRIO_USBH_ISR
};

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_STACKPTR int     _StackMain[2048/sizeof(int)];
static OS_TASK             _TCBMain;

static OS_STACKPTR int     _StackIsr[1536/sizeof(int)];
static OS_TASK             _TCBIsr;

static USBH_NOTIFICATION_HOOK _Hook;

static const char *EPType[] = { "CONTROL", "ISO", "BULK", "INT" };

/*********************************************************************
*
*       Static Code
*
**********************************************************************
*/

/*********************************************************************
*
*       _GetPortSpeed
*
*  Function description
*    Returns the given USBH speed as a string.
*/
static const char * _GetPortSpeed(USBH_SPEED Speed) {
  switch (Speed) {
  case USBH_LOW_SPEED:
    return "LowSpeed";
  case USBH_FULL_SPEED:
    return "FullSpeed";
  case USBH_HIGH_SPEED:
    return "HighSpeed";
  default:
    break;
  }
  return "Unknown";
}


/*********************************************************************
*
*       _cbOnAddRemoveDevice
*
*  Function description
*    Callback, called when a device is added or removed.
*    Called in the context of the USBH_Task.
*    The functionality in this routine should not block!
*/
static void _cbOnAddRemoveDevice(void * pContext, U8 DevIndex, USBH_DEVICE_EVENT Event) {
  USBH_BULK_HANDLE      hDevice;
  USBH_BULK_DEVICE_INFO DevInfo;
  USBH_BULK_EP_INFO     EndpointInfo;
  U32                   NumBytes;
  USBH_STATUS           Status;
  unsigned              i;
  U8                   *p;
  U8                    Buff[50];

  (void)pContext;
  switch (Event) {
  case USBH_DEVICE_EVENT_ADD:
    USBH_Logf_Application("**** New Device connected [%d]", DevIndex);
    hDevice = USBH_BULK_Open(DevIndex);
    if (hDevice == 0) {
      break;
    }
    //
    // Print device info.
    //
    Status = USBH_BULK_GetDeviceInfo(hDevice, &DevInfo);
    if (Status != USBH_STATUS_SUCCESS) {
      USBH_Logf_Application("USBH_BULK_GetDeviceInfo failed, 0x%x", Status);
    } else {
      USBH_Logf_Application("   Vendor  Id = 0x%0.4X", DevInfo.VendorId);
      USBH_Logf_Application("   Product Id = 0x%0.4X", DevInfo.ProductId);
      USBH_Logf_Application("   Interface  = %u", DevInfo.InterfaceNo);
      USBH_BULK_GetSerialNumber(hDevice, sizeof(Buff) - 1, Buff, &NumBytes);
      p = Buff;
      for (i = 0; i < NumBytes; i++) {
        if (Buff[i] != 0) {
          *p++ = Buff[i];
        }
      }
      *p = 0;
      USBH_Logf_Application("   Serial no. = %s", Buff);
      USBH_Logf_Application("   Speed      = %s", _GetPortSpeed(DevInfo.Speed));
      USBH_Logf_Application("   Endpoints:");
      for (i = 0; i < DevInfo.NumEPs; i++) {
        if (USBH_BULK_GetEndpointInfo(hDevice, i, &EndpointInfo) != USBH_STATUS_SUCCESS) {
          USBH_Logf_Application("USBH_BULK_GetEndpointInfo() failed");
          break;
        }
        USBH_Logf_Application("      Addr %02x, Type = %s", EndpointInfo.Addr, EPType[EndpointInfo.Type]);
      }
    }
    USBH_BULK_Close(hDevice);
    break;
  case USBH_DEVICE_EVENT_REMOVE:
    USBH_Logf_Application("**** Device removed [%d]", DevIndex);
    break;
  default:;   // Should never happen
  }
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/
/*********************************************************************
*
*       MainTask
*/
#ifdef __cplusplus
extern "C" {     /* Make sure we have C-declarations in C++ programs */
#endif
void MainTask(void);
#ifdef __cplusplus
}
#endif
void MainTask(void) {
  USBH_INTERFACE_MASK InterfaceMask;

  USBH_Init();
  OS_SetPriority(OS_GetTaskID(), TASK_PRIO_APP);                                       // This task has the lowest prio for real-time application.
                                                                                       // Tasks using emUSB-Host API should always have a lower priority than emUSB-Host main and ISR tasks.
  OS_CREATETASK(&_TCBMain, "USBH_Task", USBH_Task, TASK_PRIO_USBH_MAIN, _StackMain);   // Start USBH main task
  OS_CREATETASK(&_TCBIsr, "USBH_isr", USBH_ISRTask, TASK_PRIO_USBH_ISR, _StackIsr);    // Start USBH ISR task
  USBH_BULK_Init(NULL);
  USBH_MEMSET(&InterfaceMask, 0, sizeof(USBH_INTERFACE_MASK)); // Empty mask means all devices.
  USBH_BULK_AddNotification(&_Hook, _cbOnAddRemoveDevice, NULL, &InterfaceMask);
  while (1) {
    BSP_ToggleLED(1);
    OS_Delay(100);
  }
}

/*************************** End of file ****************************/
