/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 2003 - 2019     SEGGER Microcontroller GmbH              *
*                                                                    *
*       www.segger.com     Support: www.segger.com/ticket            *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host * USB Host stack for embedded applications        *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product.                          *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host version: V2.16                                    *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------

File    : USBH_MIDI_HID_Keyboard.c
Purpose : This sample uses an attached keyboard to transmit
          MIDI events to all MIDI devices connected to the host.
*/

/*********************************************************************
*
*       #include section
*
**********************************************************************
*/

#include "RTOS.h"
#include "BSP.h"
#include "USBH.h"
#include "USBH_HID.h"
#include "USBH_MIDI.h"

/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/

#define MAX_DATA_ITEMS        10

/*********************************************************************
*
*       Defines, non-configurable
*
**********************************************************************
*/

/*********************************************************************
*
*       MIDI note numbers
*/
#define MIDI_NOTE_F4           65
#define MIDI_NOTE_E4           64
#define MIDI_NOTE_D4_SHARP     63
#define MIDI_NOTE_D4           62
#define MIDI_NOTE_C4_SHARP     61
#define MIDI_NOTE_C4           60
#define MIDI_NOTE_B3           59
#define MIDI_NOTE_A3_SHARP     58
#define MIDI_NOTE_A3           57
#define MIDI_NOTE_G3_SHARP     56
#define MIDI_NOTE_G3           55
#define MIDI_NOTE_F3_SHARP     54
#define MIDI_NOTE_F3           53
#define MIDI_NOTE_E3           52
#define MIDI_NOTE_D3_SHARP     51
#define MIDI_NOTE_D3           50
#define MIDI_NOTE_C3_SHARP     49
#define MIDI_NOTE_C3           48

/*********************************************************************
*
*       Local types
*
**********************************************************************
*/

enum {
  TASK_PRIO_APP = 150,
  TASK_PRIO_USBH_MAIN,
  TASK_PRIO_USBH_ISR
};

typedef struct {
  USBH_DEVICE_EVENT Event;
  U8                DevIndex;
} PNP_EVENT;

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_STACKPTR int        _StackMain[1536/sizeof(int)];
static OS_TASK                _TCBMain;
static OS_STACKPTR int        _StackIsr[1276/sizeof(int)];
static OS_TASK                _TCBIsr;
static OS_MAILBOX             _HIDMailBox;
static USBH_HID_KEYBOARD_DATA _aHIDEvents[MAX_DATA_ITEMS];
static OS_MAILBOX             _PnPMailbox;
static PNP_EVENT              _aPnPEvents[32];
static USBH_MIDI_HANDLE       _aMidiDevice[16];
static USBH_NOTIFICATION_HOOK _MidiHook;
static unsigned               _Channel;
static unsigned               _Cable;

/*********************************************************************
*
*       Static Code
*
**********************************************************************
*/

/*********************************************************************
*
*       _HID_KeyboardEvent
*
*  Function description
*    Called from the USBH task when a keyboard event occurs.
*
*  Parameters
*    pKeyData - Pointer to keyboard event data.
*/
static void _HID_KeyboardEvent(USBH_HID_KEYBOARD_DATA * pKeyData) {
  OS_PutMailCond(&_HIDMailBox, pKeyData);
}

/*********************************************************************
*
*       _HID_DeviceEvent
*
*  Function description
*    Called when a HID device is added or removed.
*
*  Parameters
*    pContext - Pointer to user-provided context.
*    DevIndex - Zero-based device index.
*    Event    - Reason for callback.
*
*  Additional information
*    Called in the context of the USBH task.
*/
static void _HID_DeviceEvent(void * pContext, U8 DevIndex, USBH_DEVICE_EVENT Event) {
  //
  (void)pContext;
  //
  switch (Event) {
  case USBH_DEVICE_EVENT_ADD:
    USBH_Logf_Application("HID Device %d added", DevIndex);
    break;
  case USBH_DEVICE_EVENT_REMOVE:
    USBH_Logf_Application("HID Device %d removed", DevIndex);
    break;
  default:
    break;   // Should never happen
  }
}

/*********************************************************************
*
*       _MIDI_DeviceEvent()
*
*  Function description
*    Called when a device is added or removed.
*
*  Parameters
*    pContext - Pointer to user-provided context.
*    DevIndex - Zero-based device index.
*    Event    - Reason for callback.
*
*  Additional information
*    Called in the context of the USBH task.
*/
static void _MIDI_DeviceEvent(void * pContext, U8 DevIndex, USBH_DEVICE_EVENT Event) {
  PNP_EVENT PnP;
  //
  (void)pContext;
  //
  PnP.Event    = Event;
  PnP.DevIndex = DevIndex;
  //
  OS_PutMail(&_PnPMailbox, &PnP);
}

/*********************************************************************
*
*       _SendNote()
*
*  Function description
*    Send MIDI notes to all channels.
*
*  Parameters
*    Note      - MIDI note to send.
*    pHidEvent - Pointer to HID event data.
*/
static void _SendNote(unsigned Note, const USBH_HID_KEYBOARD_DATA *pHidEvent) {
  unsigned Velocity;
  unsigned i;
  U32      Event;
  //
  Velocity = pHidEvent->Value ? 64 : 32;
  Event   = (_Cable << 28) +
            (pHidEvent->Value ? 0x09900000uL : 0x08800000uL) +
            ((_Channel) << 16) +
            (Note << 8) +
            Velocity;
  //
  USBH_Logf_Application("HID: %2u %s     MIDI: %02x %02x %02x %02x",
                        pHidEvent->Code,
                        pHidEvent->Value ? "Down" : "Up  ",
                        (Event >> 24) & 0xFFu,
                        (Event >> 16) & 0xFFu,
                        (Event >> 8) & 0xFFu,
                        (Event >> 0) & 0xFFu);
  //
  for (i = 0; i < SEGGER_COUNTOF(_aMidiDevice); ++i) {
    if (_aMidiDevice[i] != USBH_MIDI_INVALID_HANDLE) {
      USBH_MIDI_WrEvent(_aMidiDevice[i], Event);
    }
  }
}

/*********************************************************************
*
*       _SelectChannel()
*
*  Function description
*    Select MIDI channel to send events on.
*
*  Parameters
*    Channel - Zero-based MIDI channel.
*/
static void _SelectChannel(unsigned Channel) {
  _Channel = Channel;
  USBH_Logf_Application("Select MIDI Channel %u", Channel+1);
}

/*********************************************************************
*
*       _AddDevice()
*
*  Function description
*    Add device to local control.
*
*  Parameters
*    pEvent - Pointer to plug event.
*/
static void _AddDevice(PNP_EVENT *pEvent) {
  USBH_MIDI_DEVICE_INFO DevInfo;
  USBH_STATUS           Status;
  //
  if (_aMidiDevice[pEvent->DevIndex] != USBH_MIDI_INVALID_HANDLE) {
    USBH_Warnf_Application("MIDI device %d already open!", pEvent->DevIndex);
  } else {
    _aMidiDevice[pEvent->DevIndex] = USBH_MIDI_Open(pEvent->DevIndex);
    if (_aMidiDevice[pEvent->DevIndex] != USBH_MIDI_INVALID_HANDLE) {
      Status = USBH_MIDI_GetDeviceInfo(_aMidiDevice[pEvent->DevIndex], &DevInfo);
      if (Status == USBH_STATUS_SUCCESS) {
        USBH_Logf_Application("MIDI device %d added: %u In cables, %u Out cables", pEvent->DevIndex, DevInfo.NumInCables, DevInfo.NumOutCables);
      } else {
        USBH_Warnf_Application("Error getting MIDI device %d information!", pEvent->DevIndex);
      }
    } else {
      USBH_Warnf_Application("Error opening MIDI device %d!", pEvent->DevIndex);
    }
  }
}

/*********************************************************************
*
*       _RemoveDevice()
*
*  Function description
*    Remove device from local control.
*
*  Parameters
*    pEvent - Pointer to unplug event.
*/
static void _RemoveDevice(PNP_EVENT *pEvent) {
  if (_aMidiDevice[pEvent->DevIndex] == USBH_MIDI_INVALID_HANDLE) {
    USBH_Warnf_Application("MIDI device %d already removed!", pEvent->DevIndex);
  } else {
    (void)USBH_MIDI_Close(_aMidiDevice[pEvent->DevIndex]);
    _aMidiDevice[pEvent->DevIndex] = USBH_MIDI_INVALID_HANDLE;
    USBH_Logf_Application("MIDI device %d removed", pEvent->DevIndex);
  }
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

#ifdef __cplusplus
extern "C" {     /* Make sure we have C-declarations in C++ programs */
#endif
void MainTask(void);
#ifdef __cplusplus
}
#endif

/*********************************************************************
*
*       MainTask()
*
*  Function description
*    Start all keyboard tasks.
*/
void MainTask(void) {
  USBH_HID_KEYBOARD_DATA HidEvent;
  PNP_EVENT              PnPEvent;
  //
  USBH_Init();
  OS_SetPriority(OS_GetTaskID(), TASK_PRIO_APP);                                          // This task has the lowest prio for real-time application.
  //                                                                                      // Tasks using emUSB-Host API should always have a lower priority than emUSB-Host main and ISR tasks.
  OS_CREATETASK(&_TCBMain, "USBH_Task", USBH_Task,    TASK_PRIO_USBH_MAIN, _StackMain);   // Start USBH main task
  OS_CREATETASK(&_TCBIsr,  "USBH_isr",  USBH_ISRTask, TASK_PRIO_USBH_ISR,  _StackIsr);    // Start USBH ISR task
  //
  USBH_HID_Init();
  USBH_HID_SetOnKeyboardStateChange(_HID_KeyboardEvent);
  USBH_HID_RegisterNotification(_HID_DeviceEvent, NULL);
  //
  // Some USB host controllers lack the support for multiple USB transfers at once.
  // Only one operation at a time is allowed.
  // This the case for the KINETIS FS driver with the Kinetis USBOTG (USB0) controller.
  // If you are using a different controller you can comment the following line to
  // enable LED update functionality for keyboards.
  //
  USBH_HID_ConfigureAllowLEDUpdate(0);
  //
  // Add MIDI device support.
  //
  USBH_MIDI_Init();
  USBH_MIDI_AddNotification(&_MidiHook, _MIDI_DeviceEvent, NULL);
  //
  // Create mailboxes to store the HID and PnP events.
  //
  OS_CREATEMB(&_HIDMailBox, sizeof(USBH_HID_KEYBOARD_DATA), SEGGER_COUNTOF(_aHIDEvents), &_aHIDEvents);
  OS_CREATEMB(&_PnPMailbox, sizeof(PNP_EVENT),              SEGGER_COUNTOF(_aPnPEvents), &_aPnPEvents);
  //
  for (;;) {
    //
    OS_Delay(10);
    //
    while (OS_GetMailCond(&_PnPMailbox, &PnPEvent) == 0) {
      if (PnPEvent.Event == USBH_DEVICE_EVENT_ADD) {
        _AddDevice(&PnPEvent);
      } else {
        _RemoveDevice(&PnPEvent);
      }
    }
    //
    if (OS_GetMailCond(&_HIDMailBox, &HidEvent) == 0) {
      BSP_ToggleLED(1);
      //
      switch (HidEvent.Code) {
      //
      case  4 /*A*/: _SendNote(MIDI_NOTE_C3,       &HidEvent); break;
      case 26 /*W*/: _SendNote(MIDI_NOTE_C3_SHARP, &HidEvent); break;
      case 22 /*S*/: _SendNote(MIDI_NOTE_D3,       &HidEvent); break;
      case  8 /*E*/: _SendNote(MIDI_NOTE_D3_SHARP, &HidEvent); break;
      case  7 /*D*/: _SendNote(MIDI_NOTE_E3,       &HidEvent); break;
      case  9 /*F*/: _SendNote(MIDI_NOTE_F3,       &HidEvent); break;
      case 23 /*T*/: _SendNote(MIDI_NOTE_F3_SHARP, &HidEvent); break;
      case 10 /*G*/: _SendNote(MIDI_NOTE_G3,       &HidEvent); break;
      case 28 /*Y*/: _SendNote(MIDI_NOTE_G3_SHARP, &HidEvent); break;
      case 11 /*H*/: _SendNote(MIDI_NOTE_A3,       &HidEvent); break;
      case 24 /*U*/: _SendNote(MIDI_NOTE_A3_SHARP, &HidEvent); break;
      case 13 /*J*/: _SendNote(MIDI_NOTE_B3,       &HidEvent); break;
      case 14 /*K*/: _SendNote(MIDI_NOTE_C4,       &HidEvent); break;
      case 18 /*O*/: _SendNote(MIDI_NOTE_C4_SHARP, &HidEvent); break;
      case 15 /*L*/: _SendNote(MIDI_NOTE_D4,       &HidEvent); break;
      case 19 /*P*/: _SendNote(MIDI_NOTE_D4_SHARP, &HidEvent); break;
      case 51 /*;*/: _SendNote(MIDI_NOTE_E4,       &HidEvent); break;
      case 52 /*'*/: _SendNote(MIDI_NOTE_F4,       &HidEvent); break;
      //
      case 58 /*F1*/:
      case 59 /*F2*/:
      case 60 /*F3*/:
      case 61 /*F4*/:
      case 62 /*F5*/:
      case 63 /*F6*/:
      case 64 /*F7*/:
      case 65 /*F8*/:
      case 66 /*F9*/:
      case 67 /*F10*/:
      case 68 /*F11*/:
      case 69 /*F12*/:
        if (HidEvent.Value) {
          _SelectChannel(HidEvent.Code - 58);
        }
        break;
      //
      default:
        break;
      }
    }
  }
}

/*************************** End of file ****************************/
