/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 2003 - 2019     SEGGER Microcontroller GmbH              *
*                                                                    *
*       www.segger.com     Support: www.segger.com/ticket            *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host * USB Host stack for embedded applications        *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product.                          *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host version: V2.16                                    *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------

File    : USBH_MIDI_Message.c
Purpose : Scroll a message across Novation Launchpad Mini MK2.

Additional information:

  Preparation:
    Plug in a Launchpad Mini Mk2 to the USB bus.

  Expected behavior:
    A message scrolls across the display.
*/

/*********************************************************************
*
*       #include section
*
**********************************************************************
*/

#include "RTOS.h"
#include "BSP.h"
#include "USBH.h"
#include "USBH_MIDI.h"
#include "SEGGER.h"
#include <stdio.h>
#include <stdlib.h>

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/

/*********************************************************************
*
*       Event decomposition
*/
#define MIDI_GET_CABLE(E)    ((E) >> 28)
#define MIDI_GET_CID(E)      (((E) >> 24) & 0xF)
#define MIDI_GET_STATUS(E)   (((E) >> 16) & 0xFF)
#define MIDI_GET_CHANNEL(E)  (((E) >> 16) & 0xF)
#define MIDI_GET_DATA2(E)    (unsigned)(((E) >> 8) & 0x7Fu)
#define MIDI_GET_DATA3(E)    (unsigned)((E) & 0x7Fu)
#define MIDI_GET_DATAX(E)    ((MIDI_GET_DATA3(E) << 7) + MIDI_GET_DATA2(E))

/*********************************************************************
*
*       Event composition
*/
#define MIDI_PUT_CABLE(E)    (((E) & 0xF) << 28)
#define MIDI_PUT_CID(E)      (((E) & 0xF) << 24)
#define MIDI_PUT_STATUS(E)   (((E) & 0xFF) << 16)
#define MIDI_PUT_CHANNEL(E)  (((E) & 0xF) << 16)
#define MIDI_PUT_DATA2(E)    (((E) & 0x7Fu) << 8)
#define MIDI_PUT_DATA3(E)    ((E) & 0x7Fu)
#define MIDI_PUT_DATAX(E)    (MIDI_PUT_DATA3((E) >> 7) + MIDI_PUT_DATA2(E))

/*********************************************************************
*
*       Launchpad LED colors
*/
#define RED(L)         (L)
#define GREEN(L)       ((L) << 4)
#define ORANGE(L)      (GREEN(3) | RED(3))

/*********************************************************************
*
*       To see glyphs easier.
*/
#define ________ 0x00
#define _______X 0x01
#define ______X_ 0x02
#define ______XX 0x03
#define _____X__ 0x04
#define _____X_X 0x05
#define _____XX_ 0x06
#define _____XXX 0x07
#define ____X___ 0x08
#define ____X__X 0x09
#define ____X_X_ 0x0A
#define ____X_XX 0x0B
#define ____XX__ 0x0C
#define ____XX_X 0x0D
#define ____XXX_ 0x0E
#define ____XXXX 0x0F
#define ___X____ 0x10
#define ___X___X 0x11
#define ___X__X_ 0x12
#define ___X__XX 0x13
#define ___X_X__ 0x14
#define ___X_X_X 0x15
#define ___X_XX_ 0x16
#define ___X_XXX 0x17
#define ___XX___ 0x18
#define ___XX__X 0x19
#define ___XX_X_ 0x1A
#define ___XX_XX 0x1B
#define ___XXX__ 0x1C
#define ___XXX_X 0x1D
#define ___XXXX_ 0x1E
#define ___XXXXX 0x1F
#define __X_____ 0x20
#define __X____X 0x21
#define __X___X_ 0x22
#define __X___XX 0x23
#define __X__X__ 0x24
#define __X__X_X 0x25
#define __X__XX_ 0x26
#define __X__XXX 0x27
#define __X_X___ 0x28
#define __X_X__X 0x29
#define __X_X_X_ 0x2A
#define __X_X_XX 0x2B
#define __X_XX__ 0x2C
#define __X_XX_X 0x2D
#define __X_XXX_ 0x2E
#define __X_XXXX 0x2F
#define __XX____ 0x30
#define __XX___X 0x31
#define __XX__X_ 0x32
#define __XX__XX 0x33
#define __XX_X__ 0x34
#define __XX_X_X 0x35
#define __XX_XX_ 0x36
#define __XX_XXX 0x37
#define __XXX___ 0x38
#define __XXX__X 0x39
#define __XXX_X_ 0x3A
#define __XXX_XX 0x3B
#define __XXXX__ 0x3C
#define __XXXX_X 0x3D
#define __XXXXX_ 0x3E
#define __XXXXXX 0x3F
#define _X______ 0x40
#define _X_____X 0x41
#define _X____X_ 0x42
#define _X____XX 0x43
#define _X___X__ 0x44
#define _X___X_X 0x45
#define _X___XX_ 0x46
#define _X___XXX 0x47
#define _X__X___ 0x48
#define _X__X__X 0x49
#define _X__X_X_ 0x4A
#define _X__X_XX 0x4B
#define _X__XX__ 0x4C
#define _X__XX_X 0x4D
#define _X__XXX_ 0x4E
#define _X__XXXX 0x4F
#define _X_X____ 0x50
#define _X_X___X 0x51
#define _X_X__X_ 0x52
#define _X_X__XX 0x53
#define _X_X_X__ 0x54
#define _X_X_X_X 0x55
#define _X_X_XX_ 0x56
#define _X_X_XXX 0x57
#define _X_XX___ 0x58
#define _X_XX__X 0x59
#define _X_XX_X_ 0x5A
#define _X_XX_XX 0x5B
#define _X_XXX__ 0x5C
#define _X_XXX_X 0x5D
#define _X_XXXX_ 0x5E
#define _X_XXXXX 0x5F
#define _XX_____ 0x60
#define _XX____X 0x61
#define _XX___X_ 0x62
#define _XX___XX 0x63
#define _XX__X__ 0x64
#define _XX__X_X 0x65
#define _XX__XX_ 0x66
#define _XX__XXX 0x67
#define _XX_X___ 0x68
#define _XX_X__X 0x69
#define _XX_X_X_ 0x6A
#define _XX_X_XX 0x6B
#define _XX_XX__ 0x6C
#define _XX_XX_X 0x6D
#define _XX_XXX_ 0x6E
#define _XX_XXXX 0x6F
#define _XXX____ 0x70
#define _XXX___X 0x71
#define _XXX__X_ 0x72
#define _XXX__XX 0x73
#define _XXX_X__ 0x74
#define _XXX_X_X 0x75
#define _XXX_XX_ 0x76
#define _XXX_XXX 0x77
#define _XXXX___ 0x78
#define _XXXX__X 0x79
#define _XXXX_X_ 0x7A
#define _XXXX_XX 0x7B
#define _XXXXX__ 0x7C
#define _XXXXX_X 0x7D
#define _XXXXXX_ 0x7E
#define _XXXXXXX 0x7F
#define X_______ 0x80
#define X______X 0x81
#define X_____X_ 0x82
#define X_____XX 0x83
#define X____X__ 0x84
#define X____X_X 0x85
#define X____XX_ 0x86
#define X____XXX 0x87
#define X___X___ 0x88
#define X___X__X 0x89
#define X___X_X_ 0x8A
#define X___X_XX 0x8B
#define X___XX__ 0x8C
#define X___XX_X 0x8D
#define X___XXX_ 0x8E
#define X___XXXX 0x8F
#define X__X____ 0x90
#define X__X___X 0x91
#define X__X__X_ 0x92
#define X__X__XX 0x93
#define X__X_X__ 0x94
#define X__X_X_X 0x95
#define X__X_XX_ 0x96
#define X__X_XXX 0x97
#define X__XX___ 0x98
#define X__XX__X 0x99
#define X__XX_X_ 0x9A
#define X__XX_XX 0x9B
#define X__XXX__ 0x9C
#define X__XXX_X 0x9D
#define X__XXXX_ 0x9E
#define X__XXXXX 0x9F
#define X_X_____ 0xA0
#define X_X____X 0xA1
#define X_X___X_ 0xA2
#define X_X___XX 0xA3
#define X_X__X__ 0xA4
#define X_X__X_X 0xA5
#define X_X__XX_ 0xA6
#define X_X__XXX 0xA7
#define X_X_X___ 0xA8
#define X_X_X__X 0xA9
#define X_X_X_X_ 0xAA
#define X_X_X_XX 0xAB
#define X_X_XX__ 0xAC
#define X_X_XX_X 0xAD
#define X_X_XXX_ 0xAE
#define X_X_XXXX 0xAF
#define X_XX____ 0xB0
#define X_XX___X 0xB1
#define X_XX__X_ 0xB2
#define X_XX__XX 0xB3
#define X_XX_X__ 0xB4
#define X_XX_X_X 0xB5
#define X_XX_XX_ 0xB6
#define X_XX_XXX 0xB7
#define X_XXX___ 0xB8
#define X_XXX__X 0xB9
#define X_XXX_X_ 0xBA
#define X_XXX_XX 0xBB
#define X_XXXX__ 0xBC
#define X_XXXX_X 0xBD
#define X_XXXXX_ 0xBE
#define X_XXXXXX 0xBF
#define XX______ 0xC0
#define XX_____X 0xC1
#define XX____X_ 0xC2
#define XX____XX 0xC3
#define XX___X__ 0xC4
#define XX___X_X 0xC5
#define XX___XX_ 0xC6
#define XX___XXX 0xC7
#define XX__X___ 0xC8
#define XX__X__X 0xC9
#define XX__X_X_ 0xCA
#define XX__X_XX 0xCB
#define XX__XX__ 0xCC
#define XX__XX_X 0xCD
#define XX__XXX_ 0xCE
#define XX__XXXX 0xCF
#define XX_X____ 0xD0
#define XX_X___X 0xD1
#define XX_X__X_ 0xD2
#define XX_X__XX 0xD3
#define XX_X_X__ 0xD4
#define XX_X_X_X 0xD5
#define XX_X_XX_ 0xD6
#define XX_X_XXX 0xD7
#define XX_XX___ 0xD8
#define XX_XX__X 0xD9
#define XX_XX_X_ 0xDA
#define XX_XX_XX 0xDB
#define XX_XXX__ 0xDC
#define XX_XXX_X 0xDD
#define XX_XXXX_ 0xDE
#define XX_XXXXX 0xDF
#define XXX_____ 0xE0
#define XXX____X 0xE1
#define XXX___X_ 0xE2
#define XXX___XX 0xE3
#define XXX__X__ 0xE4
#define XXX__X_X 0xE5
#define XXX__XX_ 0xE6
#define XXX__XXX 0xE7
#define XXX_X___ 0xE8
#define XXX_X__X 0xE9
#define XXX_X_X_ 0xEA
#define XXX_X_XX 0xEB
#define XXX_XX__ 0xEC
#define XXX_XX_X 0xED
#define XXX_XXX_ 0xEE
#define XXX_XXXX 0xEF
#define XXXX____ 0xF0
#define XXXX___X 0xF1
#define XXXX__X_ 0xF2
#define XXXX__XX 0xF3
#define XXXX_X__ 0xF4
#define XXXX_X_X 0xF5
#define XXXX_XX_ 0xF6
#define XXXX_XXX 0xF7
#define XXXXX___ 0xF8
#define XXXXX__X 0xF9
#define XXXXX_X_ 0xFA
#define XXXXX_XX 0xFB
#define XXXXXX__ 0xFC
#define XXXXXX_X 0xFD
#define XXXXXXX_ 0xFE
#define XXXXXXXX 0xFF


/*********************************************************************
*
*       Local types
*
**********************************************************************
*/

enum {
  TASK_PRIO_SCROLL = 150,
  TASK_PRIO_LIGHTS,
  TASK_PRIO_APP,
  TASK_PRIO_USBH_MAIN,
  TASK_PRIO_USBH_ISR
};

typedef struct {
  unsigned    Width;
  unsigned    Height;
  unsigned    Chars;
  const U8  * pWidths;
  const U16 * pIndex;
  const U8  * pBitmap;
} BITMAP_FONT;

typedef struct {
  int        Width;
  int        Height;
  const U8 * pRaster;
} GLYPH;

typedef struct {
  USBH_DEVICE_EVENT Event;
  U8                DevIndex;
} PNP_EVENT;

typedef enum {
  MIDI_DEVICE_STATE_CONTROLLER,
  MIDI_DEVICE_STATE_SHUTDOWN,
  MIDI_DEVICE_STATE_HALTED,
} MIDI_DEVICE_STATE;

typedef struct {
  USBH_MIDI_HANDLE           hMIDI;
  U8                         DevIndex;
  volatile MIDI_DEVICE_STATE State;
  U8                         aBuffer[64];
} MIDI_DEVICE;

/*********************************************************************
*
*       Static const data
*
**********************************************************************
*/

static const U8 _Fixed_5x7_aRaster[] = {
  ________,
  X_X_X___,
  ________,
  X___X___,
  ________,
  X_X_X___,
  ________,

  /* U+0020 'space' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  ________,
  ________,
  ________,
  ________,
  ________,

  /* U+0021 'exclam' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X_______,
  X_______,
  X_______,
  X_______,
  ________,
  X_______,
  ________,

  /* U+0022 'quotedbl' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _X_X____,
  _X_X____,
  _X_X____,
  ________,
  ________,
  ________,
  ________,

  /* U+0023 'numbersign' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  _X_X____,
  XXXXX___,
  _X_X____,
  XXXXX___,
  _X_X____,
  ________,

  /* U+0024 'dollar' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  _XXX____,
  X_X_____,
  _XXX____,
  __X_X___,
  _XXX____,
  ________,

  /* U+0025 'percent' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X_______,
  X__X____,
  __X_____,
  _X______,
  X__X____,
  ___X____,
  ________,

  /* U+0026 'ampersand' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  _X______,
  X_X_____,
  _X______,
  X_X_____,
  _X_X____,
  ________,

  /* U+0027 'quotesingle' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  __X_____,
  __X_____,
  __X_____,
  ________,
  ________,
  ________,
  ________,

  /* U+0028 'parenleft' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  __X_____,
  _X______,
  _X______,
  _X______,
  _X______,
  __X_____,
  ________,

  /* U+0029 'parenright' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _X______,
  __X_____,
  __X_____,
  __X_____,
  __X_____,
  _X______,
  ________,

  /* U+002a 'asterisk' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  _X_X____,
  __X_____,
  _XXX____,
  __X_____,
  _X_X____,
  ________,

  /* U+002b 'plus' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  __X_____,
  __X_____,
  XXXXX___,
  __X_____,
  __X_____,
  ________,

  /* U+002c 'comma' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  ________,
  ________,
  _X______,
  _X______,
  X_______,

  /* U+002d 'hyphen' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  ________,
  XXXX____,
  ________,
  ________,
  ________,

  /* U+002e 'period' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  ________,
  ________,
  _XX_____,
  _XX_____,
  ________,

  /* U+002f 'slash' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ___X____,
  __X_____,
  _X______,
  X_______,
  ________,
  ________,

  /* U+0030 'zero' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  __X_____,
  _X_X____,
  _X_X____,
  _X_X____,
  _X_X____,
  __X_____,
  ________,

  /* U+0031 'one' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  __X_____,
  _XX_____,
  __X_____,
  __X_____,
  __X_____,
  _XXX____,
  ________,

  /* U+0032 'two' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  X__X____,
  ___X____,
  __X_____,
  _X______,
  XXXX____,
  ________,

  /* U+0033 'three' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  XXXX____,
  ___X____,
  _XX_____,
  ___X____,
  X__X____,
  _XX_____,
  ________,

  /* U+0034 'four' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  __X_____,
  _XX_____,
  X_X_____,
  XXXX____,
  __X_____,
  __X_____,
  ________,

  /* U+0035 'five' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  XXXX____,
  X_______,
  XXX_____,
  ___X____,
  X__X____,
  _XX_____,
  ________,

  /* U+0036 'six' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  X_______,
  XXX_____,
  X__X____,
  X__X____,
  _XX_____,
  ________,

  /* U+0037 'seven' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  XXXX____,
  ___X____,
  __X_____,
  __X_____,
  _X______,
  _X______,
  ________,

  /* U+0038 'eight' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  X__X____,
  _XX_____,
  X__X____,
  X__X____,
  _XX_____,
  ________,

  /* U+0039 'nine' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  X__X____,
  X__X____,
  _XXX____,
  ___X____,
  _XX_____,
  ________,

  /* U+003a 'colon' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  _XX_____,
  _XX_____,
  ________,
  _XX_____,
  _XX_____,
  ________,

  /* U+003b 'semicolon' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  _XX_____,
  _XX_____,
  ________,
  _XX_____,
  _X______,
  X_______,

  /* U+003c 'less' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ___X____,
  __X_____,
  _X______,
  __X_____,
  ___X____,
  ________,

  /* U+003d 'equal' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  XXXX____,
  ________,
  XXXX____,
  ________,
  ________,

  /* U+003e 'greater' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  _X______,
  __X_____,
  ___X____,
  __X_____,
  _X______,
  ________,

  /* U+003f 'question' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  __X_____,
  _X_X____,
  ___X____,
  __X_____,
  ________,
  __X_____,
  ________,

  /* U+0040 'at' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  X__X____,
  X_XX____,
  X_XX____,
  X_______,
  _XX_____,
  ________,

  /* U+0041 'A' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  X__X____,
  X__X____,
  XXXX____,
  X__X____,
  X__X____,
  ________,

  /* U+0042 'B' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  XXX_____,
  X__X____,
  XXX_____,
  X__X____,
  X__X____,
  XXX_____,
  ________,

  /* U+0043 'C' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  X__X____,
  X_______,
  X_______,
  X__X____,
  _XX_____,
  ________,

  /* U+0044 'D' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  XXX_____,
  X__X____,
  X__X____,
  X__X____,
  X__X____,
  XXX_____,
  ________,

  /* U+0045 'E' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  XXXX____,
  X_______,
  XXX_____,
  X_______,
  X_______,
  XXXX____,
  ________,

  /* U+0046 'F' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  XXXX____,
  X_______,
  XXX_____,
  X_______,
  X_______,
  X_______,
  ________,

  /* U+0047 'G' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  X__X____,
  X_______,
  X_XX____,
  X__X____,
  _XXX____,
  ________,

  /* U+0048 'H' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X__X____,
  X__X____,
  XXXX____,
  X__X____,
  X__X____,
  X__X____,
  ________,

  /* U+0049 'I' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XXX____,
  __X_____,
  __X_____,
  __X_____,
  __X_____,
  _XXX____,
  ________,

  /* U+004a 'J' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ___X____,
  ___X____,
  ___X____,
  ___X____,
  X__X____,
  _XX_____,
  ________,

  /* U+004b 'K' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X__X____,
  X_X_____,
  XX______,
  XX______,
  X_X_____,
  X__X____,
  ________,

  /* U+004c 'L' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X_______,
  X_______,
  X_______,
  X_______,
  X_______,
  XXXX____,
  ________,

  /* U+004d 'M' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X__X____,
  XXXX____,
  XXXX____,
  X__X____,
  X__X____,
  X__X____,
  ________,

  /* U+004e 'N' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X__X____,
  XX_X____,
  XX_X____,
  X_XX____,
  X_XX____,
  X__X____,
  ________,

  /* U+004f 'O' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  X__X____,
  X__X____,
  X__X____,
  X__X____,
  _XX_____,
  ________,

  /* U+0050 'P' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  XXX_____,
  X__X____,
  X__X____,
  XXX_____,
  X_______,
  X_______,
  ________,

  /* U+0051 'Q' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  X__X____,
  X__X____,
  X__X____,
  XX_X____,
  _XX_____,
  ___X____,

  /* U+0052 'R' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  XXX_____,
  X__X____,
  X__X____,
  XXX_____,
  X_X_____,
  X__X____,
  ________,

  /* U+0053 'S' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  X__X____,
  _X______,
  __X_____,
  X__X____,
  _XX_____,
  ________,

  /* U+0054 'T' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XXX____,
  __X_____,
  __X_____,
  __X_____,
  __X_____,
  __X_____,
  ________,

  /* U+0055 'U' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X__X____,
  X__X____,
  X__X____,
  X__X____,
  X__X____,
  _XX_____,
  ________,

  /* U+0056 'V' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X__X____,
  X__X____,
  X__X____,
  X__X____,
  _XX_____,
  _XX_____,
  ________,

  /* U+0057 'W' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X__X____,
  X__X____,
  X__X____,
  XXXX____,
  XXXX____,
  X__X____,
  ________,

  /* U+0058 'X' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X__X____,
  X__X____,
  _XX_____,
  _XX_____,
  X__X____,
  X__X____,
  ________,

  /* U+0059 'Y' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _X_X____,
  _X_X____,
  _X_X____,
  __X_____,
  __X_____,
  __X_____,
  ________,

  /* U+005a 'Z' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  XXXX____,
  ___X____,
  __X_____,
  _X______,
  X_______,
  XXXX____,
  ________,

  /* U+005b 'bracketleft' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XXX____,
  _X______,
  _X______,
  _X______,
  _X______,
  _XXX____,
  ________,

  /* U+005c 'backslash' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  X_______,
  _X______,
  __X_____,
  ___X____,
  ________,
  ________,

  /* U+005d 'bracketright' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XXX____,
  ___X____,
  ___X____,
  ___X____,
  ___X____,
  _XXX____,
  ________,

  /* U+005e 'asciicircum' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  __X_____,
  _X_X____,
  ________,
  ________,
  ________,
  ________,
  ________,

  /* U+005f 'underscore' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  ________,
  ________,
  ________,
  XXXX____,
  ________,

  /* U+0060 'grave' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _X______,
  __X_____,
  ________,
  ________,
  ________,
  ________,
  ________,

  /* U+0061 'a' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  _XXX____,
  X__X____,
  X_XX____,
  _X_X____,
  ________,

  /* U+0062 'b' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X_______,
  X_______,
  XXX_____,
  X__X____,
  X__X____,
  XXX_____,
  ________,

  /* U+0063 'c' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  _XX_____,
  X_______,
  X_______,
  _XX_____,
  ________,

  /* U+0064 'd' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ___X____,
  ___X____,
  _XXX____,
  X__X____,
  X__X____,
  _XXX____,
  ________,

  /* U+0065 'e' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  _XX_____,
  X_XX____,
  XX______,
  _XX_____,
  ________,

  /* U+0066 'f' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  __X_____,
  _X_X____,
  _X______,
  XXX_____,
  _X______,
  _X______,
  ________,

  /* U+0067 'g' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  _XXX____,
  X__X____,
  _XX_____,
  X_______,
  _XXX____,

  /* U+0068 'h' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X_______,
  X_______,
  XXX_____,
  X__X____,
  X__X____,
  X__X____,
  ________,

  /* U+0069 'i' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  __X_____,
  ________,
  _XX_____,
  __X_____,
  __X_____,
  _XXX____,
  ________,

  /* U+006a 'j' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ___X____,
  ________,
  ___X____,
  ___X____,
  ___X____,
  _X_X____,
  __X_____,

  /* U+006b 'k' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  X_______,
  X_______,
  X_X_____,
  XX______,
  X_X_____,
  X__X____,
  ________,

  /* U+006c 'l' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _XX_____,
  __X_____,
  __X_____,
  __X_____,
  __X_____,
  _XXX____,
  ________,

  /* U+006d 'm' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  X_X_____,
  XXXX____,
  X__X____,
  X__X____,
  ________,

  /* U+006e 'n' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  XXX_____,
  X__X____,
  X__X____,
  X__X____,
  ________,

  /* U+006f 'o' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  _XX_____,
  X__X____,
  X__X____,
  _XX_____,
  ________,

  /* U+0070 'p' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  XXX_____,
  X__X____,
  X__X____,
  XXX_____,
  X_______,

  /* U+0071 'q' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  _XXX____,
  X__X____,
  X__X____,
  _XXX____,
  ___X____,

  /* U+0072 'r' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  XXX_____,
  X__X____,
  X_______,
  X_______,
  ________,

  /* U+0073 's' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  _XXX____,
  XX______,
  __XX____,
  XXX_____,
  ________,

  /* U+0074 't' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _X______,
  _X______,
  XXX_____,
  _X______,
  _X______,
  __XX____,
  ________,

  /* U+0075 'u' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  X__X____,
  X__X____,
  X__X____,
  _XXX____,
  ________,

  /* U+0076 'v' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  _X_X____,
  _X_X____,
  _X_X____,
  __X_____,
  ________,

  /* U+0077 'w' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  X__X____,
  X__X____,
  XXXX____,
  XXXX____,
  ________,

  /* U+0078 'x' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  X__X____,
  _XX_____,
  _XX_____,
  X__X____,
  ________,

  /* U+0079 'y' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  X__X____,
  X__X____,
  _X_X____,
  __X_____,
  _X______,

  /* U+007a 'z' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ________,
  ________,
  XXXX____,
  __X_____,
  _X______,
  XXXX____,
  ________,

  /* U+007b 'braceleft' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  ___X____,
  __X_____,
  _XX_____,
  __X_____,
  __X_____,
  ___X____,
  ________,

  /* U+007c 'bar' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  __X_____,
  __X_____,
  __X_____,
  __X_____,
  __X_____,
  __X_____,
  ________,

  /* U+007d 'braceright' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _X______,
  __X_____,
  __XX____,
  __X_____,
  __X_____,
  _X______,
  ________,

  /* U+007e 'asciitilde' width 5, bbx 0, bby -1, bbw 5, bbh 7 */
  _X_X____,
  X_X_____,
  ________,
  ________,
  ________,
  ________,
  ________,
};

static const U8 _Fixed_5x7_aWidths[] = {
  5,
  2,  // 0x20
  2,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  3,
  5,
  5,
  5,
  5,  // 0x30
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,  // 0x40
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,  // 0x50
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,  // 0x60
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
  5,
};

static const U16 _Fixed_5x7_aIndex[] = {
  0,
  32,
  33,
  34,
  35,
  36,
  37,
  38,
  39,
  40,
  41,
  42,
  43,
  44,
  45,
  46,
  47,
  48,
  49,
  50,
  51,
  52,
  53,
  54,
  55,
  56,
  57,
  58,
  59,
  60,
  61,
  62,
  63,
  64,
  65,
  66,
  67,
  68,
  69,
  70,
  71,
  72,
  73,
  74,
  75,
  76,
  77,
  78,
  79,
  80,
  81,
  82,
  83,
  84,
  85,
  86,
  87,
  88,
  89,
  90,
  91,
  92,
  93,
  94,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  102,
  103,
  104,
  105,
  106,
  107,
  108,
  109,
  110,
  111,
  112,
  113,
  114,
  115,
  116,
  117,
  118,
  119,
  120,
  121,
  122,
  123,
  124,
  125,
  126,
};

static const BITMAP_FONT _Fixed_5x7 = {
  /* .Width   = */ 5,
  /* .Height  = */ 7,
  /* .Chars   = */ 96,
  /* .pWidths = */ _Fixed_5x7_aWidths,
  /* .pIndex  = */ _Fixed_5x7_aIndex,
  /* .pBitmap = */ _Fixed_5x7_aRaster,
};

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_STACKPTR int        _StackMain[1536/sizeof(int)];
static OS_TASK                _TCBMain;
static OS_STACKPTR int        _StackIsr[1276/sizeof(int)];
static OS_TASK                _TCBIsr;
static OS_STACKPTR int        _StackScroll[1536/sizeof(int)];
static OS_TASK                _TCBScroll;
static OS_STACKPTR int        _StackLights[2048/sizeof(int)];
static OS_TASK                _TCBLights;
static MIDI_DEVICE            _aDevice[8];
static OS_MAILBOX             _PnPMailbox;
static PNP_EVENT              _aPnPEvents[32];
static USBH_NOTIFICATION_HOOK _Hook;
static volatile MIDI_DEVICE * _pLaunchpad;
static volatile MIDI_DEVICE * _pLightsInst;
static U8                     _Display[8][8];

/*********************************************************************
*
*       Static Code
*
**********************************************************************
*/

/*********************************************************************
*
*       _LightsTask()
*
*  Function description
*    Keep LEDs refreshed.
*
*  Additional information
*    This task sends LED updates to the Launchpad as Note On events.
*    When a change in the display is detected, Note On events are
*    sent to reflect the change.  The MIDI endpoint is buffered
*    in order to increase performance, with 16 Note On events
*    being sent as a group with the typical endpoint maximum packet
*    size of 64 bytes.
*/
static void _LightsTask(void) {
  U32      Event;
  unsigned R;
  unsigned C;
  U8       Color;
  U8       aPrev[8][8];
  int      Level;
  int      Count;
  int      Direction;
  //
  memset(aPrev, 0xFF, sizeof(aPrev));
  Level     = 0;
  Count     = 0;
  Direction = 1;
  //
  for (;;) {
    //
    // If the launchpad is removed, wait for it to arrive again.
    //
    if (_pLightsInst == NULL || USBH_MIDI_GetStatus(_pLightsInst->hMIDI) != USBH_STATUS_SUCCESS) {
      memset(aPrev, 0xFF, sizeof(aPrev));
      _pLightsInst = NULL;
      OS_Delay(100);
      continue;
    }
    //
    // Iterate over all tracks.
    //
    ++Count;
    if (Count == 7) {
      Count = 0;
      Level += Direction;
      if (Level < 0 || 3 < Level) {
        Direction = -Direction;
        Level += 2*Direction;
      }
    }
    //
    for (R = 0; R < 8; ++R) {
      for (C = 0; C < 8; ++C) {
        //
        Color = _Display[C][R];
        if (aPrev[C][R] != Color) {
          aPrev[C][R] = Color;
          Event  = MIDI_PUT_STATUS(0x90) + MIDI_PUT_CID(0x9);
          Event += MIDI_PUT_DATA2(R*16 + C);
          Event += MIDI_PUT_DATA3(Color);
          USBH_MIDI_WrEvent(_pLightsInst->hMIDI, Event);
        }
      }
    }
    //
    // Always flush data to device.
    //
    USBH_MIDI_Send(_pLightsInst->hMIDI);
    //
    // Refresh 200 times per second.
    //
    OS_Delay(10);
  }
}

/*********************************************************************
*
*       _MatchGlyph()
*
*  Function description
*    Compare function for bsearch.
*
*  Parameters
*    p1 - Pointer to left operand.
*    p2 - Pointer to right operand.
*
*  Return value
*    -1, 0, or +1 according to comparison.
*/
static int _MatchGlyph(const void *p1, const void *p2) {
  U16 u1;
  U16 u2;
  //
  u1 = *(U16 *)p1;
  u2 = *(U16 *)p2;
  if (u1 < u2) {
    return -1;
  } else if (u1 == u2) {
    return 0;
  } else {
    return +1;
  }
}

/*********************************************************************
*
*       _FindGlyph()
*
*  Function description
*    Find glyph in font.
*
*  Parameters
*    pFont  - Pointer to font to search.
*    Ch     - Character glyph to search for.
*    pGlyph - Pointer to object that receives the glyph data.
*
*  Return value
*    == 0 - Glyph not found.
*    != 0 - Glyph found.
*/
static int _FindGlyph(const BITMAP_FONT *pFont, int Ch, GLYPH *pGlyph) {
  unsigned   Index;
  U16      * s;
  U16        Key;
  //   
  Key = Ch;
  s = bsearch(&Key,
              pFont->pIndex,
              pFont->Chars,
              sizeof(U16),
              _MatchGlyph);
  //
  // If we didn't find glyph, cut short now.
  //
  if (s == NULL) {
    return 0;
  }
  //
  if (pGlyph) {
    //
    // Fill in information about glyph from font.
    //
    Index = (unsigned)(s - pFont->pIndex);
    //
    pGlyph->Height  = pFont->Height;
    pGlyph->Width   = pFont->pWidths[Index];
    pGlyph->pRaster = pFont->pBitmap + (pFont->Width+7)/8 * pFont->Height * Index;
  }
  //
  // We did find a glyph.
  //
  return 1;
}

/*********************************************************************
*
*       _DrawPixel()
*
*  Function description
*    Draw pixel on the LED matrix.
*
*  Parameters
*    x     - X coordinate.
*    y     - Y coordinate.
*    Color - Pixel color.
*
*  Additional information
*    The coordinates are clipped.
*/
static void _DrawPixel(int x, int y, int Color) {
  if (0 <= x && x <= 7 && 0 <= y && y <= 7) {
    _Display[x][y] = Color;
  }
}

/*********************************************************************
*
*       _DrawGlyph()
*
*  Function description
*    Draw glyph on the LED matrix.
*
*  Parameters
*    x      - X coordinate of top left corner.
*    y      - Y coordinate of top left corner.
*    pGlyph - Pointer to glyph data.
*    Color  - Pixel color.
*
*  Additional information
*    All coordinates are clipped.
*/
static void _DrawGlyph(int x, int y, const GLYPH *pGlyph, int Color) {
  int        w;
  int        h;
  const U8 * pRaster;

  pRaster = pGlyph->pRaster;
  for (h = 0; h < pGlyph->Height; ++h) {
    U8 b = 0;
    for (w = 0; w < pGlyph->Width; ++w) {
      if ((w & 7) == 0) {
        b = *pRaster++;
      }
      if (b & 0x80) {
        _DrawPixel(x+w, y+h, Color);
      }
      b <<= 1;
    }
  }
}

/*********************************************************************
*
*       _DrawStringN()
*
*  Function description
*    Draw length-limited string on the LED matrix.
*
*  Parameters
*    pFont   - Pointer to font to use.
*    Color   - Color of all pixels written.
*    x       - X coordinate of top left corner.
*    y       - Y coordinate of top left corner.
*    sText   - Pointer to (optionally) zero-terminated text string.
*    TextLen - Maximum number of characters to write.
*
*  Additional information
*    All coordinates are clipped.
*/
static void _DrawStringN(const BITMAP_FONT *pFont, int Color, int x, int y, const char *sText, unsigned TextLen) {
  GLYPH Glyph;
  //
  if (sText == NULL) {
    return;
  }
  //
  while (*sText && TextLen > 0) {
    int ch;
    //
    ch = *sText++;
    --TextLen;
    //
    // Only attempt to draw the glyph if the font has an encoding for the character.
    //
    if (_FindGlyph(pFont, ch, &Glyph)) {
      _DrawGlyph(x, y, &Glyph, Color);
      x += Glyph.Width;
    }
  }
}

/*********************************************************************
*
*       _DrawString()
*
*  Function description
*    Draw zero-terminated string on the LED matrix.
*
*  Parameters
*    pFont - Pointer to font to use.
*    Color - Color of all pixels written.
*    x     - X coordinate of top left corner.
*    y     - Y coordinate of top left corner.
*    sText - Pointer to zero-terminated text string.
*
*  Additional information
*    All coordinates are clipped.
*/
static void _DrawString(const BITMAP_FONT *pFont, int Color, int x, int y, const char *sText) {
  if (sText != NULL) {
    _DrawStringN(pFont, Color, x, y, sText, strlen(sText));
  }
}

/*********************************************************************
*
*       _CalcWidthN()
*
*  Function description
*    Calculate pixel width of length-limited string.
*
*  Parameters
*    pFont   - Pointer to font to use.
*    sText   - Pointer to (optionally) zero-terminated text string.
*    TextLen - Maximum number of characters to write.
*
*  Return value
*    Width of string in pixels.
*/
static int _CalcWidthN(const BITMAP_FONT *pFont, const char *sText, unsigned TextLen) {
  GLYPH Glyph;
  int   w;
  //
  w = 0;
  //
  if (sText == NULL) {
    return 0;
  }
  //
  while (*sText && TextLen > 0) {
    int ch;
    //
    ch = *sText++;
    --TextLen;
    //
    // Only attempt to draw the glyph if the font has an encoding for the character.
    //
    if (_FindGlyph(pFont, ch, &Glyph)) {
      w += Glyph.Width;
    }
  }
  //  
  // Number of pixels moved.
  //
  return w;
}

/*********************************************************************
*
*       _CalcWidth()
*
*  Function description
*    Calculate pixel width of zero-terminated string.
*
*  Parameters
*    pFont -  Pointer to font to use.
*    sText - Pointer to zero-terminated text string.
*
*  Return value
*    Width of string in pixels.
*/
static int _CalcWidth(const BITMAP_FONT *pFont, const char *sText) {
  if (sText == NULL) {
    return 0;
  } else {
    return _CalcWidthN(pFont, sText, strlen(sText));
  }
}

/*********************************************************************
*
*       _ScrollTask()
*
*  Function description
*    Scrolls a message across the display.
*/
static void _ScrollTask(void) {
  const char * sText;
  int          x;
  int          y;
  int          Width;
  //
  //
  while (_pLightsInst == NULL) {
    OS_Delay(50);
  }
  OS_Delay(2000);  // Wait for animation startup to finish.
  //
  sText = "SEGGER, the Embedded Experts!";
  Width = _CalcWidth(&_Fixed_5x7, sText);
  //
  for (;;) {
    x = 8;
    y = 1;
    while (x + Width > 0) {
      --x;
      _DrawString(&_Fixed_5x7, RED(2), x, y, sText);
      OS_Delay(80);
      _DrawString(&_Fixed_5x7, 0, x, y, sText);
    }
  }
}

/*********************************************************************
*
*       _MIDI_DeviceEvent()
*
*  Function description
*    Callback, called when a device is added or removed.
*
*  Parameters
*    pContext - Pointer to user-provided context.
*    DevIndex - Zero-based device index.
*    Event    - Type of event (add, remove).
*
*  Additional information
*    Called in the context of the USBH_Task.
*    The functionality in this routine should not block!
*/
static void _MIDI_DeviceEvent(void * pContext, U8 DevIndex, USBH_DEVICE_EVENT Event) {
  PNP_EVENT PnP;
  //
  (void)pContext;
  //
  PnP.Event    = Event;
  PnP.DevIndex = DevIndex;
  //
  OS_PutMail(&_PnPMailbox, &PnP);
}

/*********************************************************************
*
*       _AddDevice()
*
*  Function description
*    Add device to local control.
*
*  Parameters
*    pEvent - Pointer to plug event.
*/
static void _AddDevice(PNP_EVENT *pEvent) {
  USBH_MIDI_DEVICE_INFO   MidiInfo;
  MIDI_DEVICE           * pInst;
  //
  pInst = &_aDevice[pEvent->DevIndex];
  pInst->hMIDI = USBH_MIDI_Open(pEvent->DevIndex);
  //
  if (pInst->hMIDI != USBH_MIDI_INVALID_HANDLE) {
    //
    // Assign a buffer to speed up I/O.
    //
    USBH_MIDI_SetBuffer(pInst->hMIDI, &pInst->aBuffer[0], sizeof(pInst->aBuffer));
    //
    // Check for a control surface or rendering surface.
    //
    USBH_MIDI_GetDeviceInfo(pInst->hMIDI, &MidiInfo);
    if (MidiInfo.VendorId == 0x1235 && MidiInfo.ProductId == 0x0036) {
      //
      // Launchpad Mini MK2 device added, use it as the control surface.
      // Create tasks associated with it.
      //
      pInst->State = MIDI_DEVICE_STATE_CONTROLLER;
      _pLaunchpad  = pInst;
      _pLightsInst = pInst;
      //
      USBH_Logf_Application("**** Launchpad Mini MK2 MIDI device %d added and started for control", pInst->DevIndex);
    } else {
      USBH_Logf_Application("**** MIDI device %d ignored", pEvent->DevIndex);
      USBH_MIDI_Close(pInst->hMIDI);
      pInst->hMIDI = USBH_MIDI_INVALID_HANDLE;
    }
  } else {
      USBH_Logf_Application("**** MIDI device %d cannot be opened", pEvent->DevIndex);
  }
}

/*********************************************************************
*
*       _RemoveDevice()
*
*  Function description
*    Remove device from local control.
*
*  Parameters
*    pEvent - Pointer to unplug event.
*/
static void _RemoveDevice(PNP_EVENT *pEvent) {
  MIDI_DEVICE * pInst;
  //
  pInst = &_aDevice[pEvent->DevIndex];
  //
  if (pInst->hMIDI != USBH_MIDI_INVALID_HANDLE) {
    if (pInst == _pLaunchpad) {
      //
      // Wait for lights and pads tasks to finish using device.
      //
      while (_pLightsInst != NULL) {
        OS_Delay(50);
      }
    } else {
      //
      // A device we're rendering notes to: ask it to stop rendering.
      //
      pInst->State = MIDI_DEVICE_STATE_SHUTDOWN;
      while (pInst->State == MIDI_DEVICE_STATE_SHUTDOWN) {
        OS_Delay(50);
      }
    }
    //
    (void)USBH_MIDI_Close(pInst->hMIDI);
    //
    pInst->hMIDI = USBH_MIDI_INVALID_HANDLE;
  }
  USBH_Logf_Application("**** MIDI device %d stopped and removed", pEvent->DevIndex);
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/*********************************************************************
*
*       MainTask
*/
#ifdef __cplusplus
extern "C" {     /* Make sure we have C-declarations in C++ programs */
#endif
void MainTask(void);
#ifdef __cplusplus
}
#endif
void MainTask(void) {
  PNP_EVENT Event;
  //
  USBH_Init();
  USBH_MIDI_Init();
  USBH_MIDI_AddNotification(&_Hook, _MIDI_DeviceEvent, NULL);
  //
  OS_CREATEMB(&_PnPMailbox, sizeof(PNP_EVENT), SEGGER_COUNTOF(_aPnPEvents), &_aPnPEvents);
  //
  // Tasks using emUSB-Host API should always have a lower priority than emUSB-Host main and ISR tasks.
  //
  OS_SetPriority(OS_GetTaskID(), TASK_PRIO_APP);
  OS_CREATETASK(&_TCBMain,    "USBH_Task", USBH_Task,    TASK_PRIO_USBH_MAIN, _StackMain);
  OS_CREATETASK(&_TCBIsr,     "USBH_ISR",  USBH_ISRTask, TASK_PRIO_USBH_ISR,  _StackIsr);
  OS_CREATETASK(&_TCBLights,  "Lights",    _LightsTask,  TASK_PRIO_LIGHTS,    _StackLights);
  OS_CREATETASK(&_TCBScroll,  "Scroll",    _ScrollTask,  TASK_PRIO_SCROLL,    _StackScroll);
  //
  for (;;) {
    BSP_ToggleLED(1);
    //
    OS_GetMail(&_PnPMailbox, &Event);
    //
    if (Event.Event == USBH_DEVICE_EVENT_ADD) {
      _AddDevice(&Event);
    } else {
      _RemoveDevice(&Event);
    }
  }
}

/*************************** End of file ****************************/
