/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 2003 - 2019     SEGGER Microcontroller GmbH              *
*                                                                    *
*       www.segger.com     Support: www.segger.com/ticket            *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host * USB Host stack for embedded applications        *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product.                          *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host version: V2.16                                    *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------

File    : USBH_MIDI_Reversi.c
Purpose : This sample plays Reversi on the Novation Launchpad Mini MK2.

Additional information:

  Gameplay unasahmedly stolen from the emWin Reversi demo.

  Preparation:
    Plug in a Launchpad Mini Mk2 to the USB bus.

  Expected behavior:
    The game is prepared with two red and two green stones.  You
    play red.  Your valid moves are shown by pulsating red squares.
    If red wins, all red's stones pulse; if green wins, all green's
    stones pulse; if it's a draw, red and green's stones pulse
    anti-phase.

*/

/*********************************************************************
*
*       #include section
*
**********************************************************************
*/

#include "RTOS.h"
#include "BSP.h"
#include "USBH.h"
#include "USBH_MIDI.h"
#include "SEGGER.h"
#include <stdio.h>

static void _StartNewGame(void);

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/

/*********************************************************************
*
*       Event decomposition
*/
#define MIDI_GET_CABLE(E)    ((E) >> 28)
#define MIDI_GET_CID(E)      (((E) >> 24) & 0xF)
#define MIDI_GET_STATUS(E)   (((E) >> 16) & 0xFF)
#define MIDI_GET_CHANNEL(E)  (((E) >> 16) & 0xF)
#define MIDI_GET_DATA2(E)    (unsigned)(((E) >> 8) & 0x7Fu)
#define MIDI_GET_DATA3(E)    (unsigned)((E) & 0x7Fu)
#define MIDI_GET_DATAX(E)    ((MIDI_GET_DATA3(E) << 7) + MIDI_GET_DATA2(E))

/*********************************************************************
*
*       Event composition
*/
#define MIDI_PUT_CABLE(E)    (((E) & 0xF) << 28)
#define MIDI_PUT_CID(E)      (((E) & 0xF) << 24)
#define MIDI_PUT_STATUS(E)   (((E) & 0xFF) << 16)
#define MIDI_PUT_CHANNEL(E)  (((E) & 0xF) << 16)
#define MIDI_PUT_DATA2(E)    (((E) & 0x7Fu) << 8)
#define MIDI_PUT_DATA3(E)    ((E) & 0x7Fu)
#define MIDI_PUT_DATAX(E)    (MIDI_PUT_DATA3((E) >> 7) + MIDI_PUT_DATA2(E))

/*********************************************************************
*
*       Launchpad LED colors
*/
#define RED(L)         (L)
#define GREEN(L)       ((L) << 4)
#define ORANGE(L)      (GREEN(3) | RED(3))

/*********************************************************************
*
*       Board size
*/
#define NUM_CELL_X        8
#define NUM_CELL_Y        8

/*********************************************************************
*
*       Local types
*
**********************************************************************
*/

enum {
  TASK_PRIO_GAME = 150,
  TASK_PRIO_LIGHTS,
  TASK_PRIO_PADS,
  TASK_PRIO_APP,
  TASK_PRIO_USBH_MAIN,
  TASK_PRIO_USBH_ISR
};

typedef struct {
  USBH_DEVICE_EVENT Event;
  U8                DevIndex;
} PNP_EVENT;

typedef enum {
  MIDI_DEVICE_STATE_CONTROLLER,
  MIDI_DEVICE_STATE_SHUTDOWN,
  MIDI_DEVICE_STATE_HALTED,
} MIDI_DEVICE_STATE;

typedef struct {
  USBH_MIDI_HANDLE           hMIDI;
  U8                         DevIndex;
  volatile MIDI_DEVICE_STATE State;
  U8                         aBuffer[64];
} MIDI_DEVICE;

//
// Number pads 0-63 from top left of matrix to bottom right
//
typedef enum {
  PAD_DOWN,
  PAD_UP
} PAD_EVENT_TYPE;

typedef struct {
  unsigned       Pad;
  PAD_EVENT_TYPE Event;
} PAD_EVENT;

typedef struct {
  U8  aCells[NUM_CELL_X][NUM_CELL_Y];
  U8  aMoves[NUM_CELL_X][NUM_CELL_Y];
  int ActPlayer;
} BOARD;

typedef char REVERSI_AI_Func(const BOARD* pBoard, int* px, int* py);

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_STACKPTR int        _StackMain[1536/sizeof(int)];
static OS_TASK                _TCBMain;
static OS_STACKPTR int        _StackIsr[1276/sizeof(int)];
static OS_TASK                _TCBIsr;
static OS_STACKPTR int        _StackGame[1536/sizeof(int)];
static OS_TASK                _TCBGame;
static OS_STACKPTR int        _StackLights[2048/sizeof(int)];
static OS_TASK                _TCBLights;
static OS_STACKPTR int        _StackPads[2048/sizeof(int)];
static OS_TASK                _TCBPads;
static MIDI_DEVICE            _aDevice[8];
static OS_MAILBOX             _PnPMailbox;
static PNP_EVENT              _aPnPEvents[32];
static OS_MAILBOX             _PadMailbox;
static PAD_EVENT              _aPadEvents[32];
static USBH_NOTIFICATION_HOOK _Hook;
static volatile MIDI_DEVICE * _pLaunchpad;
static volatile MIDI_DEVICE * _pPadsInst;
static volatile MIDI_DEVICE * _pLightsInst;
static BOARD                  _Board;
static volatile int           _Winner;
static REVERSI_AI_Func      * _pPlayerAI[2];

/*********************************************************************
*
*       Static Code
*
**********************************************************************
*/

/*********************************************************************
*
*       _GetStone
*/
static char _GetStone(const BOARD* pBoard, int x, int y) {
  char r = 0;
  if ((x >= 0) && (y >= 0) && (x < NUM_CELL_X) && (y < NUM_CELL_Y)) {
    r = pBoard->aCells[x][y];
  }
  return r;
}

/*********************************************************************
*
*       _SetStone
*/
static void _SetStone(BOARD *pBoard, int x, int y) {
  if ((x >= 0) && (y >= 0) && (x < NUM_CELL_X) && (y < NUM_CELL_Y)) {
    pBoard->aCells[x][y] = pBoard->ActPlayer;
  }
}

/*********************************************************************
*
*       _IsValidMove
*/
static char _IsValidMove(BOARD *pBoard, int x, int y) {
  char r = 0;
  if ((x >= 0) && (y >= 0) && (x < NUM_CELL_X) && (y < NUM_CELL_Y)) {
    r = ((pBoard->aMoves[x][y]) ? 1 : 0);
  }
  return r;
}

/*********************************************************************
*
*       _CheckDirection
*/
static char _CheckDirection(const BOARD* pBoard, int x, int y, int dx, int dy) {
  char Cell;
  x += dx;
  y += dy;
  Cell = _GetStone(pBoard, x, y);
  if ((Cell != pBoard->ActPlayer) && (Cell != 0)) {
    do {
      x += dx;
      y += dy;
      Cell = _GetStone(pBoard, x, y);
    } while ((Cell != pBoard->ActPlayer) && (Cell != 0));
    return ((Cell == pBoard->ActPlayer) ? 1 : 0);
  }
  return 0;
}

/*********************************************************************
*
*       _CalcValidMoves
*/
static int _CalcValidMoves(BOARD* pBoard) {
  int x, y, r = 0;
  U8 Valid;
  for (y = 0; y < NUM_CELL_Y; y++) {
    for (x = 0; x < NUM_CELL_X; x++) {
      Valid = 0;
      if (pBoard->aCells[x][y] == 0) {
        Valid |= _CheckDirection(pBoard, x, y, -1, -1) << 0;
        Valid |= _CheckDirection(pBoard, x, y,  0, -1) << 1;
        Valid |= _CheckDirection(pBoard, x, y,  1, -1) << 2;
        Valid |= _CheckDirection(pBoard, x, y,  1,  0) << 3;
        Valid |= _CheckDirection(pBoard, x, y,  1,  1) << 4;
        Valid |= _CheckDirection(pBoard, x, y,  0,  1) << 5;
        Valid |= _CheckDirection(pBoard, x, y, -1,  1) << 6;
        Valid |= _CheckDirection(pBoard, x, y, -1,  0) << 7;
        if (Valid) {
          r++;
        }
      }
      if (Valid != pBoard->aMoves[x][y]) {
        pBoard->aMoves[x][y] = Valid;
      }
    }
  }
  return r;
}

/*********************************************************************
*
*       _DoDirection
*/
static void _DoDirection(BOARD* pBoard, int x, int y, int dx, int dy) {
  do {
    _SetStone(pBoard, x, y);
    x += dx;
    y += dy;
  } while (_GetStone(pBoard, x, y) != pBoard->ActPlayer);
}

/*********************************************************************
*
*       _MakeMove
*/
static void _MakeMove(BOARD* pBoard, int x, int y) {
  U8 Valid;
  //
  _SetStone(pBoard, x, y);
  //
  Valid = pBoard->aMoves[x][y];
  if (Valid & (U8)(1 << 0)) { _DoDirection(pBoard, x, y, -1, -1); }
  if (Valid & (U8)(1 << 1)) { _DoDirection(pBoard, x, y,  0, -1); }
  if (Valid & (U8)(1 << 2)) { _DoDirection(pBoard, x, y,  1, -1); }
  if (Valid & (U8)(1 << 3)) { _DoDirection(pBoard, x, y,  1,  0); }
  if (Valid & (U8)(1 << 4)) { _DoDirection(pBoard, x, y,  1,  1); }
  if (Valid & (U8)(1 << 5)) { _DoDirection(pBoard, x, y,  0,  1); }
  if (Valid & (U8)(1 << 6)) { _DoDirection(pBoard, x, y, -1,  1); }
  if (Valid & (U8)(1 << 7)) { _DoDirection(pBoard, x, y, -1,  0); }
}

/*********************************************************************
*
*       _CalcScore
*/
static int _CalcScore(const BOARD* pBoard) {
  int x, y, r = 0;
  char Cell;
  for (y = 0; y < NUM_CELL_Y; y++) {
    for (x = 0; x < NUM_CELL_X; x++) {
      Cell = pBoard->aCells[x][y];
      if (Cell) {
        r += (Cell == pBoard->ActPlayer) ? (1) : (-1);
      }
    }
  }
  return r;
}

/*********************************************************************
*
*       _PlayerAI_FirstValid
*/
static char _PlayerAI_FirstValid(const BOARD* pBoard, int* px, int* py) {
  int x, y;
  for (y = 0; y < NUM_CELL_Y; y++) {
    for (x = 0; x < NUM_CELL_X; x++) {
      if (pBoard->aMoves[x][y]) {
        goto End;
      }
    }
  }
End:
  *px = x;
  *py = y;
  return 1;
}

/*********************************************************************
*
*       _SetPlayer
*/
static void _SetPlayer(int Player) {
  int Score;
  //
  _Board.ActPlayer = Player;
  //
  if (!_CalcValidMoves(&_Board)) {
    int ValidMoves;
    //
    _Board.ActPlayer = 3 - Player;
    ValidMoves = _CalcValidMoves(&_Board);
    //
    _Board.ActPlayer = Player;
    _CalcValidMoves(&_Board);
    //
    if (ValidMoves) {
      //
      // No valid moves, player must pass
      //
      _SetPlayer(3 - _Board.ActPlayer);
    } else {
      //
      // No valid moves for all players, game is over
      //
      _Board.ActPlayer = 1;
      Score = _CalcScore(&_Board);
      if (Score > 0) {
        _Winner = 1;
      } else if (Score) {
        _Winner = 2;
      } else {
        _Winner = 3;
      }
    }
  }
}

/*********************************************************************
*
*       _NextPlayer
*/
static void _NextPlayer(void) {
  int x, y;
  char DoMove;
  //
  do {
    _SetPlayer(3 - _Board.ActPlayer);
    if (_pPlayerAI[_Board.ActPlayer - 1] != NULL && _Winner == 0) {
      DoMove = (*_pPlayerAI[_Board.ActPlayer - 1])(&_Board, &x, &y);
      if (DoMove) {
        _MakeMove(&_Board, x, y);
      }
    }
  } while (_pPlayerAI[_Board.ActPlayer - 1] && _Winner == 0);
}

/*********************************************************************
*
*       _GameTask()
*
*  Function description
*    Received pad events and updates pattern.
*/
static void _GameTask(void) {
  PAD_EVENT PadEvent;
  unsigned  y;
  unsigned  x;
  //
  _StartNewGame();
  //
  for (;;) {
    OS_GetMail(&_PadMailbox, &PadEvent);
    //
    if (PadEvent.Event == PAD_DOWN) {
      if (_Winner > 0) {
        _StartNewGame();
      } else {
        y = PadEvent.Pad / 8;
        x = PadEvent.Pad % 8;
        //
        if (_IsValidMove(&_Board, x, y)) {
          _MakeMove(&_Board, x, y);
          _NextPlayer();
        }
      }
    }
  }
}

/*********************************************************************
*
*       _StartNewGame
*/
static void _StartNewGame(void) {
  //
  memset(&_Board, 0, sizeof(BOARD));
  _Board.aCells[3][3] = 1;
  _Board.aCells[3][4] = 2;
  _Board.aCells[4][3] = 2;
  _Board.aCells[4][4] = 1;
  //
  _Winner = 0;
  _pPlayerAI[1] = &_PlayerAI_FirstValid;
  _SetPlayer(1);
}

/*********************************************************************
*
*       _PadsTask()
*
*  Function description
*    Receive pad events, decode, and distribute.
*
*  Additional information
*    Waits for incoming MIDI events and posts them to the pad
*    event mailbox.
*/
static void _PadsTask(void) {
  U32      Event;
  unsigned Status;
  unsigned Data2;
  unsigned Data3;
  unsigned Row;
  unsigned Col;
  //
  for (;;) {
    //
    // If the launchpad is removed, wait for it to arrive again.
    //
    if (_pPadsInst == NULL || USBH_MIDI_GetStatus(_pPadsInst->hMIDI) != USBH_STATUS_SUCCESS) {
      _pPadsInst = NULL;
      OS_Delay(100);
      continue;
    }
    //
    switch (USBH_MIDI_RdEvent(_pLaunchpad->hMIDI, &Event)) {
      //
    case USBH_STATUS_SUCCESS:
      //
      // Decode.
      //
      Status = MIDI_GET_STATUS(Event);
      Data2  = MIDI_GET_DATA2(Event);
      Data3  = MIDI_GET_DATA3(Event);
      //
      if (Status == 0x90) {
        //
        // Get absolute row and column of key pressed.
        //
        Row = Data2 >> 4;
        Col = Data2 & 0xF;
        //
        // Do not process keys at the side
        //
        if (Col <= 7) {
          PAD_EVENT PadEvent;
          //
          PadEvent.Event = Data3 > 0 ? PAD_DOWN : PAD_UP;
          PadEvent.Pad   = 8*Row + Col;
          OS_PutMail(&_PadMailbox, &PadEvent);
        }
      }
      break;
      //
    case USBH_STATUS_TIMEOUT:
      break;
      //
    case USBH_STATUS_DEVICE_REMOVED:
    case USBH_STATUS_NOT_OPENED:
      USBH_Logf_Application("MIDI device removed...");
      break;
      //
    default:
      break;
    }
  }
}

/*********************************************************************
*
*       _LightsTask()
*
*  Function description
*    Keep LEDs refreshed.
*
*  Additional information
*    This task sends LED updates to the Launchpad as Note On events.
*    When a change in the display is detected, Note On events are
*    sent to reflect the change.  The MIDI endpoint is buffered
*    in order to increase performance, with 16 Note On events
*    being sent as a group with the typical endpoint maximum packet
*    size of 64 bytes.
*/
static void _LightsTask(void) {
  U32      Event;
  unsigned R;
  unsigned C;
  U8       Color;
  U8       aPrev[8][8];
  int      Level;
  int      Count;
  int      Direction;
  //
  memset(aPrev, 0xFF, sizeof(aPrev));
  Level     = 0;
  Count     = 0;
  Direction = 1;
  //
  for (;;) {
    //
    // If the launchpad is removed, wait for it to arrive again.
    //
    if (_pLightsInst == NULL || USBH_MIDI_GetStatus(_pLightsInst->hMIDI) != USBH_STATUS_SUCCESS) {
      memset(aPrev, 0xFF, sizeof(aPrev));
      _pLightsInst = NULL;
      OS_Delay(100);
      continue;
    }
    //
    // Iterate over all tracks.
    //
    ++Count;
    if (Count == 7) {
      Count = 0;
      Level += Direction;
      if (Level < 0 || 3 < Level) {
        Direction = -Direction;
        Level += 2*Direction;
      }
    }
    //
    for (R = 0; R < 8; ++R) {
      for (C = 0; C < 8; ++C) {
        //
        // Choose LED color.
        //
        Color = 0;
        if (_Board.aCells[C][R] == 1) {
          Color = RED(3);
        } else if (_Board.aCells[C][R] == 2) {
          Color = GREEN(3);
        } else if (_Winner == 0) {
          Color = _IsValidMove(&_Board, C, R) ? RED(Level) : 0;
        }
        //
        if (_Winner != 0) {
          if (_Board.aCells[C][R] & _Winner) {
            if (_Board.aCells[C][R] == 1) {
              Color = RED(Level);
            } else {
              Color = GREEN(3-Level);
            }
          }
        }
        //
        // If there is a change of color, send a Note On to change.
        //
        if (aPrev[C][R] != Color) {
          aPrev[C][R] = Color;
          Event  = MIDI_PUT_STATUS(0x90) + MIDI_PUT_CID(0x9);
          Event += MIDI_PUT_DATA2(R*16 + C);
          Event += MIDI_PUT_DATA3(Color);
          USBH_MIDI_WrEvent(_pLightsInst->hMIDI, Event);
        }
      }
    }
    //
    // Always flush data to device.
    //
    USBH_MIDI_Send(_pLightsInst->hMIDI);
    //
    // Refresh 100 times per second.
    //
    OS_Delay(10);
  }
}

/*********************************************************************
*
*       _MIDI_DeviceEvent()
*
*  Function description
*    Callback, called when a device is added or removed.
*
*  Parameters
*    pContext - Pointer to user-provided context.
*    DevIndex - Zero-based device index.
*    Event    - Type of event (add, remove).
*
*  Additional information
*    Called in the context of the USBH_Task.
*    The functionality in this routine should not block!
*/
static void _MIDI_DeviceEvent(void * pContext, U8 DevIndex, USBH_DEVICE_EVENT Event) {
  PNP_EVENT PnP;
  //
  (void)pContext;
  //
  PnP.Event    = Event;
  PnP.DevIndex = DevIndex;
  //
  OS_PutMail(&_PnPMailbox, &PnP);
}

/*********************************************************************
*
*       _AddDevice()
*
*  Function description
*    Add device to local control.
*
*  Parameters
*    pEvent - Pointer to plug event.
*/
static void _AddDevice(PNP_EVENT *pEvent) {
  USBH_MIDI_DEVICE_INFO   MidiInfo;
  MIDI_DEVICE           * pInst;
  //
  pInst = &_aDevice[pEvent->DevIndex];
  pInst->hMIDI = USBH_MIDI_Open(pEvent->DevIndex);
  //
  if (pInst->hMIDI != USBH_MIDI_INVALID_HANDLE) {
    //
    // Assign a buffer to speed up I/O.
    //
    USBH_MIDI_SetBuffer(pInst->hMIDI, &pInst->aBuffer[0], sizeof(pInst->aBuffer));
    //
    // Check for a control surface or rendering surface.
    //
    USBH_MIDI_GetDeviceInfo(pInst->hMIDI, &MidiInfo);
    if (MidiInfo.VendorId == 0x1235 && MidiInfo.ProductId == 0x0036) {
      //
      // Launchpad Mini MK2 device added, use it as the control surface.
      // Create tasks associated with it.
      //
      pInst->State = MIDI_DEVICE_STATE_CONTROLLER;
      _pLaunchpad  = pInst;
      _pPadsInst   = pInst;
      _pLightsInst = pInst;
      //
      USBH_Logf_Application("**** Launchpad Mini MK2 MIDI device %d added and started for control", pInst->DevIndex);
    } else {
      USBH_Logf_Application("**** MIDI device %d ignored", pEvent->DevIndex);
      USBH_MIDI_Close(pInst->hMIDI);
      pInst->hMIDI = USBH_MIDI_INVALID_HANDLE;
    }
  } else {
      USBH_Logf_Application("**** MIDI device %d cannot be opened", pEvent->DevIndex);
  }
}

/*********************************************************************
*
*       _RemoveDevice()
*
*  Function description
*    Remove device from local control.
*
*  Parameters
*    pEvent - Pointer to unplug event.
*/
static void _RemoveDevice(PNP_EVENT *pEvent) {
  MIDI_DEVICE * pInst;
  //
  pInst = &_aDevice[pEvent->DevIndex];
  //
  if (pInst->hMIDI != USBH_MIDI_INVALID_HANDLE) {
    if (pInst == _pLaunchpad) {
      //
      // Wait for lights and pads tasks to finish using device.
      //
      while (_pLightsInst != NULL && _pPadsInst != NULL) {
        OS_Delay(50);
      }
    } else {
      //
      // A device we're rendering notes to: ask it to stop rendering.
      //
      pInst->State = MIDI_DEVICE_STATE_SHUTDOWN;
      while (pInst->State == MIDI_DEVICE_STATE_SHUTDOWN) {
        OS_Delay(50);
      }
    }
    //
    (void)USBH_MIDI_Close(pInst->hMIDI);
    //
    pInst->hMIDI = USBH_MIDI_INVALID_HANDLE;
  }
  USBH_Logf_Application("**** MIDI device %d stopped and removed", pEvent->DevIndex);
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/*********************************************************************
*
*       MainTask
*/
#ifdef __cplusplus
extern "C" {     /* Make sure we have C-declarations in C++ programs */
#endif
void MainTask(void);
#ifdef __cplusplus
}
#endif
void MainTask(void) {
  PNP_EVENT Event;
  //
  USBH_Init();
  USBH_MIDI_Init();
  USBH_MIDI_AddNotification(&_Hook, _MIDI_DeviceEvent, NULL);
  //
  OS_CREATEMB(&_PnPMailbox, sizeof(PNP_EVENT), SEGGER_COUNTOF(_aPnPEvents), &_aPnPEvents);
  OS_CREATEMB(&_PadMailbox, sizeof(PAD_EVENT), SEGGER_COUNTOF(_aPadEvents), &_aPadEvents);
  //
  // Tasks using emUSB-Host API should always have a lower priority than emUSB-Host main and ISR tasks.
  //
  OS_SetPriority(OS_GetTaskID(), TASK_PRIO_APP);
  OS_CREATETASK(&_TCBMain,    "USBH_Task", USBH_Task,    TASK_PRIO_USBH_MAIN, _StackMain);
  OS_CREATETASK(&_TCBIsr,     "USBH_ISR",  USBH_ISRTask, TASK_PRIO_USBH_ISR,  _StackIsr);
  OS_CREATETASK(&_TCBPads,    "Pads",      _PadsTask,    TASK_PRIO_PADS,      _StackPads);
  OS_CREATETASK(&_TCBLights,  "Lights",    _LightsTask,  TASK_PRIO_LIGHTS,    _StackLights);
  OS_CREATETASK(&_TCBGame,    "Game",      _GameTask,    TASK_PRIO_GAME,      _StackGame);
  //
  for (;;) {
    BSP_ToggleLED(1);
    //
    OS_GetMail(&_PnPMailbox, &Event);
    //
    if (Event.Event == USBH_DEVICE_EVENT_ADD) {
      _AddDevice(&Event);
    } else {
      _RemoveDevice(&Event);
    }
  }
}

/*************************** End of file ****************************/
