/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 2003 - 2019     SEGGER Microcontroller GmbH              *
*                                                                    *
*       www.segger.com     Support: www.segger.com/ticket            *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host * USB Host stack for embedded applications        *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product.                          *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host version: V2.16                                    *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------

File    : USBH_MIDI_Sqeuencer.c
Purpose : This sample is designed to present emUSBH's capability to
          enumerate MIDIStreaming devices and convert a Novation
          Launchpad Mini MK2 into a simple standalone sequencer.

Additional information:

  Preparation:
    Plug in a Launchpad Mini Mk2 to the USB bus.  Plug in a drum
    machine or synthesizer that is configured to receive MIDI
    events on MIDI channel 10.

  Expected behavior:
    An initial drum pattern is shown on the display along with
    a running 120 bpm playhead.  Use the pads to turn on and
    off 1/16th notes.  A measure runs across two rows, 8 1/16th
    notes per row.  The four parts are configured top to bottom
    as hi hat, snare, crash, and kick.

    When running, the pads trigger MIDI events to the connected
    synthesizer.
*/

/*********************************************************************
*
*       #include section
*
**********************************************************************
*/

#include "RTOS.h"
#include "BSP.h"
#include "USBH.h"
#include "USBH_MIDI.h"
#include "SEGGER.h"

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/

/*********************************************************************
*
*       Event decomposition
*/
#define MIDI_GET_CABLE(E)    ((E) >> 28)
#define MIDI_GET_CID(E)      (((E) >> 24) & 0xF)
#define MIDI_GET_STATUS(E)   (((E) >> 16) & 0xFF)
#define MIDI_GET_CHANNEL(E)  (((E) >> 16) & 0xF)
#define MIDI_GET_DATA2(E)    (unsigned)(((E) >> 8) & 0x7Fu)
#define MIDI_GET_DATA3(E)    (unsigned)((E) & 0x7Fu)
#define MIDI_GET_DATAX(E)    ((MIDI_GET_DATA3(E) << 7) + MIDI_GET_DATA2(E))

/*********************************************************************
*
*       Event composition
*/
#define MIDI_PUT_CABLE(E)    (((E) & 0xF) << 28)
#define MIDI_PUT_CID(E)      (((E) & 0xF) << 24)
#define MIDI_PUT_STATUS(E)   (((E) & 0xFF) << 16)
#define MIDI_PUT_CHANNEL(E)  (((E) & 0xF) << 16)
#define MIDI_PUT_DATA2(E)    (((E) & 0x7Fu) << 8)
#define MIDI_PUT_DATA3(E)    ((E) & 0x7Fu)
#define MIDI_PUT_DATAX(E)    (MIDI_PUT_DATA3((E) >> 7) + MIDI_PUT_DATA2(E))

/*********************************************************************
*
*       Launchpad LED colors
*/
#define RED(L)         (L)
#define GREEN(L)       ((L) << 4)
#define ORANGE(L)      (GREEN(3) | RED(3))

/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/

#define PLAYHEAD_COLOR GREEN(3)
#define NOTE_COLOR     RED(3)
#define ACCENT_COLOR   ORANGE(3)
#define PULSE_COLOR    RED(3)
#define BEAT_COLOR     GREEN(3)
#define EMPTY_COLOR    0

/*********************************************************************
*
*       Local data definitions
*
**********************************************************************
*/

enum {
  TASK_PRIO_APP = 150,
  TASK_PRIO_MIDI_RECV,
  TASK_PRIO_USBH_MAIN,
  TASK_PRIO_USBH_BEAT,
  TASK_PRIO_USBH_ISR
};

typedef struct {
  USBH_DEVICE_EVENT Event;
  U8                DevIndex;
} PNP_EVENT;

typedef enum {
  MIDI_DEVICE_STATE_CONTROLLER,
  MIDI_DEVICE_STATE_RENDER,
  MIDI_DEVICE_STATE_SHUTDOWN,
  MIDI_DEVICE_STATE_HALTED,
} MIDI_DEVICE_STATE;

typedef struct {
  USBH_MIDI_HANDLE           hMIDI;
  U8                         DevIndex;
  volatile MIDI_DEVICE_STATE State;
  OS_TASK                    RxTask;
  OS_TASK                    TxTask;
  OS_STACKPTR int            aRxStack[1500/sizeof(int)];
  OS_STACKPTR int            aTxStack[1500/sizeof(int)];
  U8                         aBuffer[64];
} MIDI_DEVICE;

typedef U8 DRUM_PATTERN[16];

//
// Number pads 0-63 from top left of matrix to bottom right
//
typedef enum {
  PAD_DOWN,
  PAD_HOLD,
  PAD_UP
} PAD_EVENT_TYPE;

typedef struct {
  unsigned       Pad;
  PAD_EVENT_TYPE Event;
} PAD_EVENT;

/*********************************************************************
*
*       Static const data
*
**********************************************************************
*/

static const U8 _aVoice[4] = {
  42,   // Closed hi-hat
  40,   // Electric snare
  49,   // Crash 1
  36    // Bass drum 1
};

static const U8 _aAccentVoice[4] = {
  46,   // Open hi-hat
  39,   // Hand clap
  57,   // Crash 2
  41    // Low floor tom
};

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_STACKPTR int        _StackMain[1536/sizeof(int)];
static OS_TASK                _TCBMain;
static OS_STACKPTR int        _StackIsr[1276/sizeof(int)];
static OS_TASK                _TCBIsr;
static OS_STACKPTR int        _StackBeat[1536/sizeof(int)];
static OS_TASK                _TCBBeat;
static OS_STACKPTR int        _StackPads[1536/sizeof(int)];
static OS_TASK                _TCBPads;
static OS_STACKPTR int        _StackControl[1536/sizeof(int)];
static OS_TASK                _TCBControl;
static OS_STACKPTR int        _StackLights[1536/sizeof(int)];
static OS_TASK                _TCBLights;
static OS_STACKPTR int        _StackPress[1536/sizeof(int)];
static OS_TASK                _TCBPress;
static MIDI_DEVICE            _aDevice[8];
static OS_MAILBOX             _PnPMailbox;
static PNP_EVENT              _aPnPEvents[32];
static OS_MAILBOX             _PadMailbox;
static PAD_EVENT              _aPadEvents[32];
static USBH_NOTIFICATION_HOOK _Hook;
static DRUM_PATTERN           _aPattern[4];
static volatile unsigned      _PlayHead;
static volatile U32           _aPadDownTime[64];
static volatile MIDI_DEVICE * _pLaunchpad;
static volatile MIDI_DEVICE * _pPadsInst;
static volatile MIDI_DEVICE * _pLightsInst;

/*********************************************************************
*
*       Static Code
*
**********************************************************************
*/

/*********************************************************************
*
*       _MIDI_DeviceEvent()
*
*  Function description
*    Callback, called when a device is added or removed.
*    Called in the context of the USBH_Task.
*    The functionality in this routine should not block!
*/
static void _MIDI_DeviceEvent(void * pContext, U8 DevIndex, USBH_DEVICE_EVENT Event) {
  PNP_EVENT PnP;
  //
  (void)pContext;
  //
  PnP.Event    = Event;
  PnP.DevIndex = DevIndex;
  //
  OS_PutMail(&_PnPMailbox, &PnP);
}

/*********************************************************************
*
*       _LightsTask()
*
*  Function description
*    Keep LEDs refreshed.
*
*  Additional information
*    This task sends LED updates to the Launchpad as Note On events.
*    When a change in the display is detected, Note On events are
*    sent to reflect the change.  The MIDI endpoint is buffered
*    in order to increase performance, with 16 Note On events
*    being sent as a group with the typical endpoint maximum packet
*    size of 64 bytes.
*/
static void _LightsTask(void) {
  U32      Event;
  unsigned R;
  unsigned C;
  U8       Color;
  U8       aPrev[4][17];
  //
  memset(aPrev, 0xFF, sizeof(aPrev));
  //
  for (;;) {
    //
    // If the launchpad is removed, wait for it to arrive again.
    //
    if (_pLightsInst == NULL || USBH_MIDI_GetStatus(_pLightsInst->hMIDI) != USBH_STATUS_SUCCESS) {
      _pLightsInst = NULL;
      OS_Delay(100);
      continue;
    }
    //
    // Iterate over all tracks.
    //
    for (R = 0; R < 4; ++R) {
      for (C = 0; C < 16; ++C) {
        //
        // Choose LED color.
        //
        if (C == _PlayHead/2) {
          Color = PLAYHEAD_COLOR;
        } else {
          Color = _aPattern[R][C];
        }
        //
        // If there is a change of color, send a Note On to change.
        //
        if (aPrev[R][C] != Color) {
          aPrev[R][C] = Color;
          Event  = MIDI_PUT_STATUS(0x90) + MIDI_PUT_CID(0x9);
          Event += MIDI_PUT_DATA2((R*32) + (C % 8) + (C / 8 * 16));
          Event += MIDI_PUT_DATA3(Color);
          USBH_MIDI_WrEvent(_pLightsInst->hMIDI, Event);
        }
      }
      //
      // Indicate the status of the track on the right-side LED.
      // The LED will illuminate for 1/32nd note according to the
      /// note at the playhead: it illuminates green for voiced
      // quarter notes and red for voiced non-quarter notes.
      //
      if (_PlayHead % 2 == 0) {
        if (_aPattern[R][_PlayHead/2] != EMPTY_COLOR) {
          if (_PlayHead/2 % 4 == 0) {
            Color = MIDI_PUT_DATA3(BEAT_COLOR);
          } else {
            Color = MIDI_PUT_DATA3(PULSE_COLOR);
          }
        } else {
          Color = MIDI_PUT_DATA3(EMPTY_COLOR);
        }
      } else {
        Color = 0;
      }
      //
      if (aPrev[R][16] != Color) {
        aPrev[R][16] = Color;
        Event  = MIDI_PUT_STATUS(0x90) + MIDI_PUT_CID(0x9);
        Event += MIDI_PUT_DATA2((R*32) + 8);
        Event += MIDI_PUT_DATA3(Color);
        USBH_MIDI_WrEvent(_pLightsInst->hMIDI, Event);
      }
    }
    //
    // Always flush data to device.
    //
    USBH_MIDI_Send(_pLightsInst->hMIDI);
    //
    // Refresh 200 times per second.
    //
    OS_Delay(10);
  }
}

/*********************************************************************
*
*       _BeatTask()
*
*  Function description
*    Keep time and dispatch notes to connected devices.
*
*  Additional information
*    This task keeps time and sends Note On events to the connected
*    MIDI devices to render the pattern.
*/
static void _BeatTask(void) {
  MIDI_DEVICE * pInst;
  U32           T0;
  U32           Event;
  unsigned      DeviceIndex;
  unsigned      TrackIndex;
  //
  T0 = OS_GetTime32();
  //
  // Run with 1/32nd note clock.
  //
  for (;;) {
    _PlayHead = (_PlayHead+1) % 32;
    //
    // 1000 ms per tick, therefore 60,000 ticks per minutes.
    // 120 1/4 notes per minute, therefore 120 * 8 = 960 1/32nds minute.
    // So, each 1/16th notes is 60,000 / 480 = 125 ms.
    //
    T0 += (1000 * 60) / (120 * 8);
    OS_DelayUntil(T0);
    //
    // Send pad data to connected devices on 16th notes.
    //
    if (_PlayHead % 2 == 0) {
      for (DeviceIndex = 0; DeviceIndex < SEGGER_COUNTOF(_aDevice); ++DeviceIndex) {
        //
        pInst = &_aDevice[DeviceIndex];
        if (pInst->State == MIDI_DEVICE_STATE_SHUTDOWN) {
          pInst->State = MIDI_DEVICE_STATE_HALTED;
        } else if (pInst->State == MIDI_DEVICE_STATE_RENDER) {
          //
          // This device must render notes.  Only output on cable #0.
          //
          for (TrackIndex = 0; TrackIndex < 4; ++TrackIndex) {
            if (_aPattern[TrackIndex][_PlayHead/2] > 0) {
              unsigned Voice;
              //
              // Send Note On for the selected track voice on channel #10.
              //
              Voice = _aPattern[TrackIndex][_PlayHead/2] == ACCENT_COLOR ? _aAccentVoice[TrackIndex] : _aVoice[TrackIndex];
              //
              Event = MIDI_PUT_CABLE(0) +
                      MIDI_PUT_STATUS(0x090) +
                      MIDI_PUT_CID(0x9) +
                      MIDI_PUT_CHANNEL(10-1) +
                      MIDI_PUT_DATA2(Voice) +
                      MIDI_PUT_DATA3(0x40);
              //
              USBH_MIDI_WrEvent(pInst->hMIDI, Event);
            }
          }
          //
          // Force and buffered data to rendering device.
          //
          USBH_MIDI_Send(pInst->hMIDI);
        }
      }
    }
  }
}

/*********************************************************************
*
*       _ControlTask()
*
*  Function description
*    Received pad events and updates pattern.
*/
static void _ControlTask(void) {
  PAD_EVENT PadEvent;
  unsigned  Row;
  unsigned  Col;
  //
  for (;;) {
    OS_GetMail(&_PadMailbox, &PadEvent);
    //
    // Translate to folded note position.
    //
    Row = PadEvent.Pad / 16;
    Col = PadEvent.Pad % 16;
    //
    if (PadEvent.Event == PAD_DOWN) {
      if (_aPattern[Row][Col] != EMPTY_COLOR) {
        _aPattern[Row][Col] = EMPTY_COLOR;
      } else {
        _aPattern[Row][Col] = NOTE_COLOR;
      }
    } else if (PadEvent.Event == PAD_HOLD) {
      _aPattern[Row][Col] = ACCENT_COLOR;
    }
  }
}

/*********************************************************************
*
*       _PressTask()
*
*  Function description
*    Senses pads being held and send pad-hold events to the
*    pad mailbox.
*/
static void _PressTask(void) {
  unsigned Pad;
  //
  for (;;) {
    //
    for (Pad = 0; Pad < SEGGER_COUNTOF(_aPadDownTime); ++Pad) {
      if (_aPadDownTime[Pad] != 0) {
        if (OS_GetTime32()-_aPadDownTime[Pad] > 500) {
          PAD_EVENT PadEvent;
          //
          _aPadDownTime[Pad] = 0;
          PadEvent.Pad   = Pad;
          PadEvent.Event = PAD_HOLD;
          OS_PutMail(&_PadMailbox, &PadEvent);
        }
      }
    }
    //
    OS_Delay(50);
  }
}

/*********************************************************************
*
*       _PadsTask()
*
*  Function description
*    Receive pad events, decode, and distribute.
*
*  Additional information
*    Waits for incoming MIDI events and posts them to the display
*    task which decodes them and updates the virtual keyboard to
*    reflect the physical key state.
*/
static void _PadsTask(void) {
  U32      Event;
  unsigned Status;
  unsigned Data2;
  unsigned Data3;
  unsigned Row;
  unsigned Col;
  //
  for (;;) {
    //
    // If the launchpad is removed, wait for it to arrive again.
    //
    if (_pPadsInst == NULL || USBH_MIDI_GetStatus(_pPadsInst->hMIDI) != USBH_STATUS_SUCCESS) {
      _pPadsInst = NULL;
      OS_Delay(100);
      continue;
    }
    //
    switch (USBH_MIDI_RdEvent(_pLaunchpad->hMIDI, &Event)) {
      //
    case USBH_STATUS_SUCCESS:
      //
      // Decode.
      //
      Status = MIDI_GET_STATUS(Event);
      Data2  = MIDI_GET_DATA2(Event);
      Data3  = MIDI_GET_DATA3(Event);
      //
      if (Status == 0x90) {
        //
        // Get absolute row and column of key pressed.
        //
        Row = Data2 >> 4;
        Col = Data2 & 0xF;
        //
        // Do not process keys at the side
        //
        if (Col <= 7) {
          PAD_EVENT PadEvent;
          //
          PadEvent.Event = Data3 > 0 ? PAD_DOWN : PAD_UP;
          PadEvent.Pad   = 8*Row + Col;
          OS_PutMail(&_PadMailbox, &PadEvent);
          //
          if (PadEvent.Event == PAD_DOWN) {
            USBH_Logf_Application("Pad %2u down", PadEvent.Pad);
            _aPadDownTime[PadEvent.Pad] = OS_GetTime32();
          } else {
            USBH_Logf_Application("Pad %2u up", PadEvent.Pad);
            _aPadDownTime[PadEvent.Pad] = 0;
          }
        }
      }
      break;
      //
    case USBH_STATUS_TIMEOUT:
      break;
      //
    case USBH_STATUS_DEVICE_REMOVED:
    case USBH_STATUS_NOT_OPENED:
      USBH_Logf_Application("MIDI device removed...");
      break;
      //
    default:
      break;
    }
  }
}

/*********************************************************************
*
*       _AddDevice()
*
*  Function description
*    Add device to local control.
*
*  Parameters
*    pEvent - Pointer to plug event.
*/
static void _AddDevice(PNP_EVENT *pEvent) {
  USBH_MIDI_DEVICE_INFO   MidiInfo;
  MIDI_DEVICE           * pInst;
  //
  pInst = &_aDevice[pEvent->DevIndex];
  pInst->hMIDI = USBH_MIDI_Open(pEvent->DevIndex);
  //
  if (pInst->hMIDI != USBH_MIDI_INVALID_HANDLE) {
    //
    // Assign a buffer to speed up I/O.
    //
    USBH_MIDI_SetBuffer(pInst->hMIDI, &pInst->aBuffer[0], sizeof(pInst->aBuffer));
    //
    // Check for a control surface or rendering surface.
    //
    USBH_MIDI_GetDeviceInfo(pInst->hMIDI, &MidiInfo);
    if (MidiInfo.VendorId == 0x1235 && MidiInfo.ProductId == 0x0036) {
      //
      // Launchpad Mini MK2 device added, use it as the control surface.
      // Create tasks associated with it.
      //
      pInst->State = MIDI_DEVICE_STATE_CONTROLLER;
      _pLaunchpad  = pInst;
      _pPadsInst   = pInst;
      _pLightsInst = pInst;
      //
      USBH_Logf_Application("**** Launchpad Mini MK2 MIDI device %d added and started for control", pInst->DevIndex);
    } else {
      //
      // Rendering surface.
      //
      pInst->State = MIDI_DEVICE_STATE_RENDER;
      USBH_Logf_Application("**** MIDI device %d added and started for rendering", pEvent->DevIndex);
    }
  } else {
      USBH_Logf_Application("**** MIDI device %d cannot be opened", pEvent->DevIndex);
  }
}

/*********************************************************************
*
*       _RemoveDevice()
*
*  Function description
*    Remove device from local control.
*
*  Parameters
*    pEvent - Pointer to unplug event.
*/
static void _RemoveDevice(PNP_EVENT *pEvent) {
  MIDI_DEVICE * pInst;
  //
  pInst = &_aDevice[pEvent->DevIndex];
  //
  if (pInst->hMIDI != USBH_MIDI_INVALID_HANDLE) {
    if (pInst == _pLaunchpad) {
      //
      // Wait for lights and pads tasks to finish using device.
      //
      while (_pLightsInst != NULL && _pPadsInst != NULL) {
        OS_Delay(50);
      }
    } else {
      //
      // A device we're rendering notes to: ask it to stop rendering.
      //
      pInst->State = MIDI_DEVICE_STATE_SHUTDOWN;
      while (pInst->State == MIDI_DEVICE_STATE_SHUTDOWN) {
        OS_Delay(50);
      }
    }
    //
    (void)USBH_MIDI_Close(pInst->hMIDI);
    //
    pInst->hMIDI = USBH_MIDI_INVALID_HANDLE;
  }
  USBH_Logf_Application("**** MIDI device %d stopped and removed", pEvent->DevIndex);
}

/*********************************************************************
*
*       _SetDefaultPattern()
*
*  Function description
*    Set the default pattern.
*/
static void _SetDefaultPattern(void) {
  unsigned  i;
  //
  for (i = 0; i < 16; ++i) {
    _aPattern[0][i] = (i % 2) == 0 ? NOTE_COLOR : EMPTY_COLOR;
  }
  for (i = 0; i < 16; ++i) {
    _aPattern[1][i] = (i % 8) == 4 ? NOTE_COLOR : EMPTY_COLOR;
  }
  for (i = 0; i < 16; ++i) {
    _aPattern[3][i] = (i % 4) == 0 ? NOTE_COLOR : EMPTY_COLOR;
  }
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

#ifdef __cplusplus
extern "C" {     /* Make sure we have C-declarations in C++ programs */
#endif
  void MainTask(void);
#ifdef __cplusplus
}
#endif

/*********************************************************************
*
*       MainTask
*/
void MainTask(void) {
  PNP_EVENT Event;
  //
  USBH_Init();
  USBH_MIDI_Init();
  USBH_MIDI_AddNotification(&_Hook, _MIDI_DeviceEvent, NULL);
  //
  OS_CREATEMB(&_PnPMailbox, sizeof(PNP_EVENT), SEGGER_COUNTOF(_aPnPEvents), &_aPnPEvents);
  OS_CREATEMB(&_PadMailbox, sizeof(PAD_EVENT), SEGGER_COUNTOF(_aPadEvents), &_aPadEvents);
  //
  // Tasks using emUSB-Host API should always have a lower priority than emUSB-Host main and ISR tasks.
  //
  OS_SetPriority(OS_GetTaskID(), TASK_PRIO_APP);
  OS_CREATETASK(&_TCBMain,    "USBH_Task", USBH_Task,    TASK_PRIO_USBH_MAIN, _StackMain);
  OS_CREATETASK(&_TCBIsr,     "USBH_ISR",  USBH_ISRTask, TASK_PRIO_USBH_ISR,  _StackIsr);
  OS_CREATETASK(&_TCBBeat,    "Beat",      _BeatTask,    TASK_PRIO_USBH_BEAT, _StackBeat);
  OS_CREATETASK(&_TCBPads,    "Pads",      _PadsTask,    TASK_PRIO_MIDI_RECV, _StackPads);
  OS_CREATETASK(&_TCBLights,  "Lights",    _LightsTask,  TASK_PRIO_MIDI_RECV, _StackLights);
  OS_CREATETASK(&_TCBControl, "Control",   _ControlTask, TASK_PRIO_MIDI_RECV, _StackControl);
  OS_CREATETASK(&_TCBPress,   "Press",     _PressTask,   TASK_PRIO_MIDI_RECV, _StackPress);
  //
  _SetDefaultPattern();
  //
  for (;;) {
    BSP_ToggleLED(1);
    //
    OS_GetMail(&_PnPMailbox, &Event);
    //
    if (Event.Event == USBH_DEVICE_EVENT_ADD) {
      _AddDevice(&Event);
    } else {
      _RemoveDevice(&Event);
    }
  }
}

/*************************** End of file ****************************/
