/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 2003 - 2019     SEGGER Microcontroller GmbH              *
*                                                                    *
*       www.segger.com     Support: www.segger.com/ticket            *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host * USB Host stack for embedded applications        *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product.                          *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host version: V2.16                                    *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------

File    : USBH_MIDI_Start.c
Purpose : This sample is designed to present emUSBH's capability to
          enumerate MIDIStreaming devices and read MIDI events from
          a MIDI controller.

Additional information:
  Preparations:
    None.

  Expected behavior:
    When a MID controller is connected and events are sent, the
    events are decoded and displayed on the terminal.
*/

/*********************************************************************
*
*       #include section
*
**********************************************************************
*/

#include "RTOS.h"
#include "BSP.h"
#include "USBH.h"
#include "USBH_MIDI.h"
#include "SEGGER.h"

/*********************************************************************
*
*       Local data definitions
*
**********************************************************************
*/
enum {
  TASK_PRIO_APP = 150,
  TASK_PRIO_MIDI_RECV,
  TASK_PRIO_USBH_MAIN,
  TASK_PRIO_USBH_ISR
};

typedef struct {
  USBH_DEVICE_EVENT Event;
  U8                DevIndex;
} PNP_EVENT;

typedef struct {
  USBH_MIDI_HANDLE hMIDI;
  OS_STACKPTR int  aStack[1500/sizeof(int)];
  U8               DevIndex;
  OS_TASK          Task;
} MIDI_DEVICE;

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_STACKPTR int        _StackMain[1536/sizeof(int)];
static OS_TASK                _TCBMain;
static OS_STACKPTR int        _StackIsr[1276/sizeof(int)];
static OS_TASK                _TCBIsr;
static MIDI_DEVICE            _aDevice[32];
static OS_MAILBOX             _PnPMailbox;
static PNP_EVENT              _aPnPEvents[32];
static USBH_NOTIFICATION_HOOK _Hook;

/*********************************************************************
*
*       Static Code
*
**********************************************************************
*/

/*********************************************************************
*
*       _OnDevNotify
*
*  Function description
*    Callback, called when a device is added or removed.
*    Called in the context of the USBH_Task.
*    The functionality in this routine should not block!
*/
static void _MIDI_DeviceEvent(void * pContext, U8 DevIndex, USBH_DEVICE_EVENT Event) {
  PNP_EVENT PnP;
  //
  (void)pContext;
  //
  PnP.Event    = Event;
  PnP.DevIndex = DevIndex;
  //
  OS_PutMail(&_PnPMailbox, &PnP);
}

/*********************************************************************
*
*       _KbRecvTask()
*
*  Function description
*    Virtual keyboard receive task.
*
*  Additional information
*    Waits for incoming MIDI events and posts them to the display
*    task which decodes them and updates the virtual keyboard to
*    reflect the physical key state.
*/
static void _KbRecvTask(void *pContext) {
  MIDI_DEVICE * pInst;
  U32           Event;
  //
  pInst = (MIDI_DEVICE *)pContext;
  //
  for (;;) {
    switch (USBH_MIDI_RdEvent(pInst->hMIDI, &Event)) {
      //
     case USBH_STATUS_SUCCESS:
       //
       // The Korg microKEY generates a continuous stream of
       // zero-events interspersed with true MIDI events generated
       // by the keyboard.  We ignore those zero events.
       //
       if (Event != 0) {
         USBH_Logf_Application("MIDI event: %08x", Event);
       }
       break;
       //
     case USBH_STATUS_DEVICE_REMOVED:
     case USBH_STATUS_NOT_OPENED:
       USBH_Logf_Application("MIDI device removed, terminating receiver task");
       OS_Terminate(NULL);
       break;
       //
     default:
       break;
    }
  }
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/*********************************************************************
*
*       MainTask
*/
#ifdef __cplusplus
extern "C" {     /* Make sure we have C-declarations in C++ programs */
#endif
void MainTask(void);
#ifdef __cplusplus
}
#endif
void MainTask(void) {
  PNP_EVENT Event;
  //
  USBH_Init();
  USBH_MIDI_Init();
  USBH_MIDI_AddNotification(&_Hook, _MIDI_DeviceEvent, NULL);
  //
  OS_CREATEMB(&_PnPMailbox, sizeof(PNP_EVENT), SEGGER_COUNTOF(_aPnPEvents), &_aPnPEvents);
  //                                                                                          // Tasks using emUSB-Host API should always have a lower priority than emUSB-Host main and ISR tasks.
  OS_SetPriority(OS_GetTaskID(), TASK_PRIO_APP);                                          // This task has the lowest prio for real-time application.
  OS_CREATETASK(&_TCBMain, "USBH_Task", USBH_Task,    TASK_PRIO_USBH_MAIN, _StackMain);   // Start USBH main task
  OS_CREATETASK(&_TCBIsr , "USBH_ISR",  USBH_ISRTask, TASK_PRIO_USBH_ISR,  _StackIsr);    // Start USBH ISR task
  //
  for (;;) {
    BSP_ToggleLED(1);
    //
    OS_GetMail(&_PnPMailbox, &Event);
    if (Event.Event == USBH_DEVICE_EVENT_ADD) {
      _aDevice[Event.DevIndex].hMIDI = USBH_MIDI_Open(Event.DevIndex);
      OS_CreateTaskEx(&_aDevice[Event.DevIndex].Task,
                      "RxMIDI",
                      TASK_PRIO_MIDI_RECV,
                      _KbRecvTask,
                      &_aDevice[Event.DevIndex].aStack[0],
                      sizeof(_aDevice[Event.DevIndex].aStack)
                      CTPARA_TIMESLICE,
                      &_aDevice[Event.DevIndex]);
      USBH_Logf_Application("**** MIDI device %d added and started", Event.DevIndex);
    } else {
      if (_aDevice[Event.DevIndex].hMIDI != USBH_MIDI_INVALID_HANDLE) {
        //
        // Wait for receiver task to terminate itself.
        //
        while (OS_TASK_IsTask(&_aDevice[Event.DevIndex].Task)) {
          OS_Delay(10);
        }
        //
        (void)USBH_MIDI_Close(_aDevice[Event.DevIndex].hMIDI);
        _aDevice[Event.DevIndex].hMIDI = USBH_MIDI_INVALID_HANDLE;
      }
      USBH_Logf_Application("**** MIDI device %d stopped and removed", Event.DevIndex);
    }
  }
}

/*************************** End of file ****************************/
