/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 2003 - 2018     SEGGER Microcontroller GmbH              *
*                                                                    *
*       www.segger.com     Support: support_emusb@segger.com         *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host * USB Host stack for embedded applications        *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product.                          *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host version: V2.14a                                   *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------

File    : USBH_POS_Keyboard.c
Purpose : This sample is designed to present emUSBH's capability to
          enumerate Human Interface Devices and handle the input
          data accordingly.
          This sample will try to enumerate a barcode scanner or
          keyboard and output the scanned or entered message
          to the terminal.
          All input characters will be shown on a POS terminal.

Additional information:
  Preparations:
    None.

  Expected behavior:
    When a keyboard is connected the entered message will be displayed in the terminal
    once ENTER was pressed.
    When a barcode scanner is connected the scanned barcode will be displayed on the
    terminal.
    When a smart card or NFC reader is connected and a card or tokens is inserted
    into the reader, the serial number of the tokens are displayed on the terminal.
    When a POS terminal is connected every information will be displayed on the terminal, too.

  Sample output:
    <...>
    00000001:040 USBH_Task    - **** Device added [0]

    00000006:881 MainTask     - The following message was entered: www.segger.com
*/

/*********************************************************************
*
*       #include section
*
**********************************************************************
*/
#include "RTOS.h"
#include "BSP.h"
#include "USBH.h"
#include "USBH_HID.h"
#include "SEGGER.h"
#include "USBH_TERM_API.h"
#include "USBH_Util.h"

/*********************************************************************
*
*       Defines configurable
*
**********************************************************************
*/
#define MAX_DATA_ITEMS        10
#define KEY_BUFFER_SIZE       10
#define MESSAGE_BUFFER_SIZE   256

#ifndef SUPPORT_CCID
  #define SUPPORT_CCID         0
#endif
#define CARD_SLOT_NO           0

#if SUPPORT_CCID
  #include "USBH_CCID.h"
#endif

/*********************************************************************
*
*       Defines non-configurable
*
**********************************************************************
*/
#define KEY_VALUE_SHIFT_LEFT        0xE1
#define KEY_VALUE_SHIFT_RIGHT       0xE5
#define BEGIN_OF_VALID_KEYS         0x04

/*********************************************************************
*
*       Local data definitions
*
**********************************************************************
*/
enum {
  TASK_PRIO_APP = 150,
  TASK_PRIO_USBH_MAIN,
  TASK_PRIO_USBH_ISR
};

typedef struct {
  U8  Char;
  U8  UpperCaseChar;
} SCANCODE_TO_CHAR;

typedef struct _KEY_BUFFER {
  U8 Code;
  U8 ShiftState;
} KEY_BUFFER;

/*********************************************************************
*
*       Static const data
*
**********************************************************************
*/
//
// This table is identical to the HID US keyboard table
// with the exception that the first four entries have been removed
// (non-visible keys). Any other key that is not on the main field of
// of the keyboard is also removed, (F-Keys, NumPad, Cursors etc.)
//
static const  SCANCODE_TO_CHAR _aScanCode2CharUS[] = {
  {'a',    'A'},                     // 0x04: Key a/A
  {'b',    'B'},                     // 0x05: Key b/B
  {'c',    'C'},                     // 0x06: Key c/C
  {'d',    'D'},                     // 0x07: Key d/D
  {'e',    'E'},                     // 0x08: Key e/E
  {'f',    'F'},                     // 0x09: Key f/F
  {'g',    'G'},                     // 0x0A: Key g/G
  {'h',    'H'},                     // 0x0B: Key h/H
  {'i',    'I'},                     // 0x0C: Key i/I
  {'j',    'J'},                     // 0x0D: Key j/J
  {'k',    'K'},                     // 0x0E: Key k/K
  {'l',    'L'},                     // 0x0F: Key l/L
  {'m',    'M'},                     // 0x10: Key m/M
  {'n',    'N'},                     // 0x11: Key n/N
  {'o',    'O'},                     // 0x12: Key o/O
  {'p',    'P'},                     // 0x13: Key p/P
  {'q',    'Q'},                     // 0x14: Key q/Q
  {'r',    'R'},                     // 0x15: Key r/R
  {'s',    'S'},                     // 0x16: Key s/S
  {'t',    'T'},                     // 0x17: Key t/T
  {'u',    'U'},                     // 0x18: Key u/U
  {'v',    'V'},                     // 0x19: Key v/V
  {'w',    'W'},                     // 0x1A: Key w/W
  {'x',    'X'},                     // 0x1B: Key x/X
  {'y',    'Y'},                     // 0x1C: Key y/Y
  {'z',    'Z'},                     // 0x1D: Key z/Z
  {'1',    '!'},                     // 0x1E: Key 1/!
  {'2',    '@'},                     // 0x1F: Key 2/@
  {'3',    '#'},                     // 0x20: Key 3/#
  {'4',    '$'},                     // 0x21: Key 4/$
  {'5',    '%'},                     // 0x22: Key 5/%
  {'6',    '^'},                     // 0x23: Key 6/^
  {'7',    '&'},                     // 0x24: Key 7/&
  {'8',    '*'},                     // 0x25: Key 8/*
  {'9',    '('},                     // 0x26: Key 9/(
  {'0',    ')'},                     // 0x27: Key 0/)
  {'\n',   '\r'},                    // 0x28: Key Return (ENTER)
  {'\x1b', '\x1b'},                  // 0x29: Key ESCAPE
  {'\b',   '\b'},                    // 0x2A: Key DELETE(Backspace)
  {'\t',   '\t'},                    // 0x2B: Key Tab
  {' ',    ' '},                     // 0x2C: Key Spacebar
  {'-',    '_'},                     // 0x2D: Key -/(underscore)
  {'=',    '+'},                     // 0x2E: Key =/+
  {'[',    '{'},                     // 0x2F: Key [/{
  {']',    '}'},                     // 0x30: Key ]/}
  {'\\',   '|'},                     // 0x31: Key \\/|
  {'#',    '~'},                     // 0x32: Key Non-US #/~
  {';',    ':'},                     // 0x33: Key ;/:
  {'\'',   '\"'},                    // 0x34: Key Apostrophe/Quotation mark
  {'`',    '~'},                     // 0x35: Key GraveAccent/Tilde
  {',',    '<'},                     // 0x36: Key,/<
  {'.',    '>'},                     // 0x37: Key ./>
  {'/',    '?'},                     // 0x38: Key //?
};

#if SUPPORT_CCID
  static const U8 _APDU_GetUID[] = { 0xFF, 0xCA, 0, 0, 0 };
#endif

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
static OS_STACKPTR int         _StackMain[1536/sizeof(int)];
static OS_TASK                 _TCBMain;
static OS_STACKPTR int         _StackIsr[1276/sizeof(int)];
static OS_TASK                 _TCBIsr;
static U8                      _ShiftActive;
static KEY_BUFFER              _aKeys[KEY_BUFFER_SIZE];
static int                     _NumKeys;
static OS_EVENT                _Event;

#if SUPPORT_CCID
  static U8                    _CCIDDevIndex;
  static I8                    _CCIDDevValid;
#endif

/*********************************************************************
*
*       Static Code
*
**********************************************************************
*/


/*********************************************************************
*
*       _OutputChar
*/
static void _OutputChar(U8 Code, U8 ShiftActive) {
  U8 Char;
  static unsigned  Line   = USBH_TERM_LINES;
  static unsigned  Pos    = 0;
  static U8        Buff[50];

  if ((Code > (SEGGER_COUNTOF(_aScanCode2CharUS) + BEGIN_OF_VALID_KEYS)) || (Code < BEGIN_OF_VALID_KEYS)) {
    return;
  }
  Code -= BEGIN_OF_VALID_KEYS;
  if (ShiftActive) {
    Char = _aScanCode2CharUS[Code].UpperCaseChar;
  } else {
    Char = _aScanCode2CharUS[Code].Char;
  }
  if (Char == '\r' || Char == '\n') {
    USBH_Logf_Application("The following message was entered: %s", Buff);
    memset(Buff, 0, sizeof(Buff));
    Line++;
    Pos = 0;
    return;
  }
  if (Char == '\b') {
    if (Pos > 0) {
      Buff[--Pos] = ' ';
    }
    USBH_TERM_Print(Line, 0, (char *)Buff);
    return;
  }
  if (Char >= ' ' && Char <= 0x7F) {
    if (Pos == 0 && Line >= USBH_TERM_LINES) {
      Line = USBH_TERM_LINES - 1;
      USBH_TERM_Scroll();
    }
    if (Pos < sizeof(Buff) - 2) {
      Buff[Pos++] = Char;
    }
    USBH_TERM_Print(Line, 0, (char *)Buff);
  }
}

/*********************************************************************
*
*       _OnKeyboardChange
*
*  Function description
*    Callback, called from the USBH task when a keyboard event occurs.
*/
static void _OnKeyboardChange(USBH_HID_KEYBOARD_DATA  * pKeyData) {
  unsigned Code;
  unsigned State;

  Code  = pKeyData->Code;
  State = pKeyData->Value;
  //
  // Check whether one of the 'Shift'-Key was pressed
  //
  if (Code == KEY_VALUE_SHIFT_LEFT || Code == KEY_VALUE_SHIFT_RIGHT) {
    _ShiftActive = State;
    return;
  }
  //
  // Store the value into key buffer,
  // the once the key is released we will store the
  // key into the message buffer
  //
  if (State) {
    _aKeys[_NumKeys].Code = Code;
    _aKeys[_NumKeys++].ShiftState = _ShiftActive;
    //
    // If we reached end of the buffer, go back to the begining
    //
    if (_NumKeys >= KEY_BUFFER_SIZE) {
      _NumKeys = 0;
    }
  } else {
    int i;
    for (i = 0; i < KEY_BUFFER_SIZE; i++) {
      if (Code == _aKeys[i].Code) {
        _OutputChar(Code, _aKeys[i].ShiftState);
        _aKeys[i].Code = 0;
        _aKeys[i].ShiftState = 0;
      }
    }
  }
}

/*********************************************************************
*
*       _OnDevNotify
*
*  Function description
*    Callback, called when a device is added or removed.
*    Call in the context of the USBH_Task.
*    The functionality in this routine should not block!
*/
static void _OnDevNotify(void * pContext, U8 DevIndex, USBH_DEVICE_EVENT Event) {
  (void)pContext;
  switch (Event) {
  case USBH_DEVICE_EVENT_ADD:
    USBH_Logf_Application("**** Device added [%d]", DevIndex);
    break;
  case USBH_DEVICE_EVENT_REMOVE:
    USBH_Logf_Application("**** Device removed [%d]", DevIndex);
    break;
  default:;   // Should never happen
  }

}
/*********************************************************************
*
*       _OnDevNotifyCCID
*
*  Function description
*    Callback, called when a device is added or removed.
*    Call in the context of the USBH_Task.
*    The functionality in this routine should not block!
*/
#if SUPPORT_CCID
static void _OnDevNotifyCCID(void * pContext, U8 DevIndex, USBH_DEVICE_EVENT Event) {
  (void)pContext;
  if (Event == USBH_DEVICE_EVENT_ADD) {
    USBH_Logf_Application("**** CCID Device added [%d]", DevIndex);
    _CCIDDevIndex = DevIndex;
    _CCIDDevValid = 1;
  } else {
    USBH_Logf_Application("**** CCID Device removed [%d]", DevIndex);
    _CCIDDevValid = 0;
  }
  OS_EVENT_Set(&_Event);
}
#endif

/*********************************************************************
*
*       _cbOnSlotChange
*
*  Function description
*    Callback, called when the status of a card slot changed.
*/
#if SUPPORT_CCID
static void _cbOnSlotChange(void *pContext, U32 SlotState) {
  USBH_USE_PARA(pContext);
  USBH_Logf_Application("New slot state %x", SlotState);
  OS_EVENT_Set(&_Event);
}
#endif


/*********************************************************************
*
*       Public code
*
**********************************************************************
*/
/*********************************************************************
*
*       MainTask
*/
#ifdef __cplusplus
extern "C" {     /* Make sure we have C-declarations in C++ programs */
#endif
void MainTask(void);
#ifdef __cplusplus
}
#endif
void MainTask(void) {
#if SUPPORT_CCID
  USBH_CCID_HANDLE Handle = USBH_CCID_INVALID_HANDLE;
  U16              SlotStatus;
  I8               CardPresent = 0;
  U32              Len;
  U8               Buff[16];
  char             aMsg[40];
  static USBH_NOTIFICATION_HOOK Hook;
  static const char HexDigit[] = "0123456789ABCDEF";
#endif

  USBH_Init();
  OS_SetPriority(OS_GetTaskID(), TASK_PRIO_APP);                                       // This task has the lowest prio for real-time application.
                                                                                       // Tasks using emUSB-Host API should always have a lower priority than emUSB-Host main and ISR tasks.
  OS_CREATETASK(&_TCBMain, "USBH_Task", USBH_Task, TASK_PRIO_USBH_MAIN, _StackMain);   // Start USBH main task
  OS_CREATETASK(&_TCBIsr, "USBH_isr", USBH_ISRTask, TASK_PRIO_USBH_ISR, _StackIsr);    // Start USBH ISR task
  OS_EVENT_CreateEx(&_Event, OS_EVENT_RESET_MODE_AUTO);
  USBH_TERM_Init();
  USBH_HID_Init();
  USBH_HID_SetOnKeyboardStateChange(_OnKeyboardChange);
  USBH_HID_RegisterNotification(_OnDevNotify, NULL);
#if SUPPORT_CCID
  USBH_CCID_Init();
  USBH_CCID_AddNotification(&Hook, _OnDevNotifyCCID, NULL);
#endif
  USBH_TERM_Print(0, 0, "          SEGGER          ");
  USBH_TERM_Print(1, -1, "It simply works!");
  for (;;) {
#if SUPPORT_CCID
    if (_CCIDDevValid == 0) {
      if (Handle != USBH_CCID_INVALID_HANDLE) {
        USBH_CCID_Close(Handle);
        Handle = USBH_CCID_INVALID_HANDLE;
      }
      OS_EVENT_WaitTimed(&_Event, 50);
      continue;
    }
    if (Handle == USBH_CCID_INVALID_HANDLE) {
      Handle = USBH_CCID_Open(_CCIDDevIndex);
      if (Handle == USBH_CCID_INVALID_HANDLE) {
        OS_Delay(300);
        continue;
      }
      USBH_CCID_SetOnSlotChange(Handle, _cbOnSlotChange, NULL);
    }
    if (USBH_CCID_GetSlotStatus(Handle, CARD_SLOT_NO, &SlotStatus) != USBH_STATUS_SUCCESS) {
      USBH_Logf_Application("Can't get slot status");
      OS_Delay(1000);
      continue;
    }
    if (CardPresent != 0) {
      if (SlotStatus != 0) {
        USBH_Logf_Application("Card removed");
        CardPresent = 0;
      }
    } else {
      if (SlotStatus == 1) {
        if (USBH_CCID_PowerOn(Handle, CARD_SLOT_NO, 0, NULL) == USBH_STATUS_SUCCESS) {
          SlotStatus = 0;
        }
      }
      if (SlotStatus == 0) {
        USBH_Logf_Application("Card inserted");
        CardPresent = 1;
        Len = sizeof(Buff);
        if (USBH_CCID_APDU(Handle, CARD_SLOT_NO, sizeof(_APDU_GetUID), _APDU_GetUID, &Len, Buff) == USBH_STATUS_SUCCESS) {
          if (Len > 2) {
            U8   *x;
            char *y;

            strcpy(aMsg, "ID ");
            Len -= 2;
            x = Buff;
            y = aMsg + 3;
            while (Len-- > 0u) {
              *y++ = HexDigit[*x >> 4];
              *y++ = HexDigit[*x++ & 0xFu];
            }
            *y = 0;
            USBH_Logf_Application(aMsg);
            USBH_TERM_Scroll();
            USBH_TERM_Print(USBH_TERM_LINES - 1, 0, aMsg);
            _OutputChar(0x28, 0);
          }
        }
      }
    }
#endif
    OS_EVENT_WaitTimed(&_Event, 500);
  }
}

/*************************** End of file ****************************/
