/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 2003 - 2018     SEGGER Microcontroller GmbH              *
*                                                                    *
*       www.segger.com     Support: support_emusb@segger.com         *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host * USB Host stack for embedded applications        *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product.                          *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host version: V2.15-r14144                             *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------

File    : USBH_TERM_API.h
Purpose : API for 2x20 POS displays
*/

/*********************************************************************
*
*       #include section
*
**********************************************************************
*/
#include <stdio.h>
#include "RTOS.h"
#include "USBH.h"
#include "USBH_CDC.h"
#include "USBH_FT232.h"
#include "USBH_TERM_API.h"

/*********************************************************************
*
*       Local defines
*
**********************************************************************
*/

typedef struct {
  U32          Mask;
  const char * sData;
  U8           Size;
  U8           ParamOff;
  U16          Delay;
} CTRL_SEQUENCE;

typedef struct {
  U8    Type;         // 0 = unused, 1 = CDC, 2 = FT232
  U8    DevIndex;
  I8    State;        // 0 = new, 1 = initialized, >1 = opened, -1 = removed
  I8    Dirty;
  U32   CtrlMask;
  union {
    USBH_CDC_HANDLE   HandleCDC;
    USBH_FT232_HANDLE Handle232;
  } u;
} TERM_INST;

typedef struct {
  U16  VID;
  U16  PID;
  U32  CtrlMask;
} KNOWN_DEVICE;

typedef struct {
  U8    Type;
  char  Name[7];
} DEV_TYPE;

/*********************************************************************
*
*       Static const data
*
**********************************************************************
*/
static CTRL_SEQUENCE _InitSequence[] = {
  { 0x0001, "\x1b@", 2, 0, 10},                               // Initialize Display
  { 0x0001, "\x02\x05\x43\x37\x03", 5, 0, 10},                // Set command mode to CD5220 -> most known command from the datasheet.
  { 0x0001, "\x1b\x2a\x04", 3, 0, 10},                        // Set Brightness Mode -> 100%
  { 0x0004, "\x1b\x60\x01", 3, 0, 10},                        // Switch cursor off
  { 0x0002, "\x1f\x43\x00", 3, 0, 10}                         // Switch cursor off

};

static CTRL_SEQUENCE _PosSequence[] = {
  { 0x0001, "\x1B\x6C\x01\x01", 4, 3, 0 },
  { 0x0002, "\x1F$\x01\x01", 4, 3, 0 },
  { 0x0004, "\x1B=00", 4, 3, 0 }
};

static KNOWN_DEVICE _KnownDevices[] = {
  { 0x076C, 0x0204, 0x0001 },               // PartnerTech CD-7220-UN POS Display
  { 0x0403, 0x6001, 0x0002 },               // Tipro POS terminal (FT232 ID)
  { 0x1208, 0x0780, 0x0002 },               // EPSON M58DC (DM-D110)
  { 0x0403, 0x6001, 0x0004 }                // Jarltech Glancetron 8034 (FT232 ID)
};

static const U16 _aVendorIds[]  = { 0x1208 };
static const U16 _aProductIds[] = { 0x0780 };

static DEV_TYPE _DevCDC   = { 1, "CDC"   };
static DEV_TYPE _DevFT232 = { 2, "FT232" };

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
static OS_STACKPTR int     _StackTerm[1536/sizeof(int)];
static OS_TASK             _TCBTerm;
static OS_RSEMA            _Mutex;
static volatile I8         _Run;

static U8                  _acData[40];

static TERM_INST           _DevList[USBH_TERM_MAX_DEVICES];
static U8                  _aScreen[USBH_TERM_LINES][USBH_TERM_VIRT_COLUMS];
static U8                  _aLen[USBH_TERM_LINES];
static U8                  _aOff[USBH_TERM_LINES];
static U32                 _LastShift;

static USBH_NOTIFICATION_HOOK CDCNotiHook;
static USBH_NOTIFICATION_HOOK FT232NotiHook;

/*********************************************************************
*
*       Static Code
*
**********************************************************************
*/
/*********************************************************************
*
*       _GetPortSpeed
*
*  Function description
*    Returns the given USBH speed as a string.
*
*/
static const char * _GetPortSpeed(USBH_SPEED Speed) {
  switch (Speed) {
  case USBH_LOW_SPEED:
    return "LowSpeed";
  case USBH_FULL_SPEED:
    return "FullSpeed";
  case USBH_HIGH_SPEED:
    return "HighSpeed";
  default:
    break;
  }
  return "Unknown";
}

/*********************************************************************
*
*       _cbOnAddRemoveDevice
*
*  Function description
*    Callback, called when a device is added or removed.
*    Call in the context of the USBH_Task.
*    The functionality in this routine should not block!
*/
static void _cbOnAddRemoveDevice(void * pContext, U8 DevIndex, USBH_DEVICE_EVENT Event) {
  unsigned i;
  DEV_TYPE *pTyp = (DEV_TYPE *)pContext;

  (void)pContext;
  switch (Event) {
  case USBH_DEVICE_EVENT_ADD:
    USBH_Logf_Application("**** %s Device added [%d]", pTyp->Name, DevIndex);
    for (i = 0; i < SEGGER_COUNTOF(_DevList); i++) {
      if (_DevList[i].Type == 0) {
        _DevList[i].DevIndex = DevIndex;
        _DevList[i].State    = 0;
        _DevList[i].Type     = pTyp->Type;
        break;
      }
    }
    break;
  case USBH_DEVICE_EVENT_REMOVE:
    USBH_Logf_Application("**** %s Device removed [%d]", pTyp->Name, DevIndex);
    for (i = 0; i < SEGGER_COUNTOF(_DevList); i++) {
      if (_DevList[i].Type == pTyp->Type && _DevList[i].DevIndex == DevIndex) {
        _DevList[i].State = -1;
        break;
      }
    }
    break;
  default:;   // Should never happen
  }
}

/*********************************************************************
*
*       _SendCtrlSeq
*
*  Function description
*    Send control sequence to the terminal.
*/
static int _SendCtrlSeq(TERM_INST * pDev, CTRL_SEQUENCE * pData, unsigned Cnt, int Param) {
  unsigned Len;

  while (Cnt-- > 0) {
    if ((pData->Mask & pDev->CtrlMask) != 0) {
      Len = pData->Size;
      if (Len > sizeof(_acData)) {
        Len = sizeof(_acData);
      }
      USBH_MEMCPY(_acData, pData->sData, Len);
      if (Param >= 0 && pData->ParamOff < Len) {
        _acData[pData->ParamOff] += Param;
      }
      if (pDev->Type == 1) {
        if (USBH_CDC_Write(pDev->u.HandleCDC, _acData, Len, NULL) != USBH_STATUS_SUCCESS) {
          USBH_Logf_Application("Failed to send data, closing...");
          pDev->State = -1;
          return 1;
        }
      }
      if (pDev->Type == 2) {
        if (USBH_FT232_Write(pDev->u.Handle232, _acData, Len, NULL) != USBH_STATUS_SUCCESS) {
          USBH_Logf_Application("Failed to send data, closing...");
          pDev->State = -1;
          return 1;
        }
      }
      if (pData->Delay != 0) {
        USBH_OS_Delay(pData->Delay);
      }
    }
    pData++;
  }
  return 0;
}

/*********************************************************************
*
*       _Shift
*
*  Function description
*    Shift line.
*/
static void _Shift(unsigned Line, U8 *pOut) {
  U8      *p;
  unsigned i;

  OS_Use(&_Mutex);
  if (_aLen[Line] < USBH_TERM_COLUMS) {
    USBH_MEMCPY(pOut, _aScreen[Line], USBH_TERM_COLUMS);
    goto End;
  }
  i = _aOff[Line];
  p = &_aScreen[Line][i];
  while (i++ < _aLen[Line]) {
    *pOut++ = *p++;
  }
  p = &_aScreen[Line][0];
  for (i = _aOff[Line]; i > 0; i--) {
    *pOut++ = *p++;
  }
End:
  OS_Unuse(&_Mutex);
}

/*********************************************************************
*
*       _Touch
*
*  Function description
*    Mark display buffer dirty.
*/
static void _Touch(void) {
  unsigned i;

  for (i = 0; i < SEGGER_COUNTOF(_DevList); i++) {
    _DevList[i].Dirty = 1;
  }
}

/*********************************************************************
*
*       _TERM_Task
*
*  Function description
*    Handles all terminals.
*/
static void _TERM_Task(void) {
  unsigned               Cnt;
  unsigned               i;
  TERM_INST            * pDev;
  U32                    Now;
  USBH_CDC_HANDLE        hCDC;
  USBH_FT232_HANDLE      h232;
  int                    Dirty;
  USBH_CDC_DEVICE_INFO   DeviceInfoCDC;
  USBH_FT232_DEVICE_INFO DeviceInfo232;
  U8                     Buff[USBH_TERM_COLUMS];

  Cnt = 0;
  pDev = _DevList;
  while (_Run != 0) {
    pDev++;
    if (Cnt-- == 0) {
      Cnt = SEGGER_COUNTOF(_DevList) - 1;
      pDev = _DevList;
      Dirty = 0;
      OS_Delay(10);
      OS_Use(&_Mutex);
      Now = USBH_OS_GetTime32();
      if (Now - _LastShift > USBH_TERM_SHIFT_DELAY) {
        _LastShift = Now;
        for (i = 0; i < USBH_TERM_LINES; i++) {
          if (_aLen[i] > USBH_TERM_COLUMS) {
            if (++_aOff[i] >= _aLen[i]) {
              _aOff[i] = 0;
            }
            Dirty = 1;
          } else {
            _aOff[i] = 0;
          }
        }
      }
      OS_Unuse(&_Mutex);
    }
    if (pDev->Type == 0) {
      continue;
    }
    if (pDev->State > 2) {
      //
      // Delay for this device necessary
      //
      pDev->State--;
      continue;
    }
    if (pDev->State == 2) {
      if (_SendCtrlSeq(pDev, _InitSequence, SEGGER_COUNTOF(_InitSequence), 0)) {
        pDev->State = -1;
      } else {
        pDev->State = 1;
      }
      continue;
    }
    if (pDev->Type == 1) {
      //
      // CDC devices
      //
      if (pDev->State == 0) {
        //
        // Initialize device
        //
        hCDC = USBH_CDC_Open(pDev->DevIndex);
        if (hCDC == 0) {
          pDev->Type = 0;
          continue;
        }
        //
        // Retrieve the information about the CDC device
        // and print them out.
        //
        USBH_CDC_GetDeviceInfo(hCDC, &DeviceInfoCDC);
        USBH_Logf_Application("Vendor  Id = 0x%0.4X", DeviceInfoCDC.VendorId);
        USBH_Logf_Application("Product Id = 0x%0.4X", DeviceInfoCDC.ProductId);
        USBH_Logf_Application("Speed      = %s", _GetPortSpeed(DeviceInfoCDC.Speed));
        //
        // Known device ?
        //
        for (i = 0; i < SEGGER_COUNTOF(_KnownDevices); i++) {
          if (_KnownDevices[i].VID == DeviceInfoCDC.VendorId &&
              _KnownDevices[i].PID == DeviceInfoCDC.ProductId) {
            break;
          }
        }
        if (i >= SEGGER_COUNTOF(_KnownDevices)) {
          //
          // Not found
          //
          USBH_Logf_Application("Unknown device, ignored");
          pDev->State = -1;
          continue;
        }
        pDev->CtrlMask =  _KnownDevices[i].CtrlMask;
        pDev->u.HandleCDC = hCDC;
        //
        // Configure the CDC device.
        //
        USBH_CDC_SetTimeouts(hCDC, 50, 50);
        USBH_CDC_AllowShortRead(hCDC, 1);
        USBH_CDC_SetCommParas(hCDC, USBH_CDC_BAUD_115200, USBH_CDC_BITS_8, USBH_CDC_STOP_BITS_1, USBH_CDC_PARITY_NONE);
        pDev->State = 50;
        pDev->Dirty = 1;
        continue;
      }
      if (pDev->State < 0) {
        USBH_CDC_Close(pDev->u.HandleCDC);
        pDev->Type = 0;
        continue;
      }
      if (pDev->Dirty != 0 || Dirty != 0) {
        //
        // Update display
        //
        pDev->Dirty = 0;
        for (i = 0; i < USBH_TERM_LINES; i++) {
          if (_SendCtrlSeq(pDev, _PosSequence, SEGGER_COUNTOF(_PosSequence), i) != 0) {
            break;
          }
          _Shift(i, Buff);
          USBH_CDC_Write(pDev->u.HandleCDC, Buff, USBH_TERM_COLUMS, NULL);
        }
      }
    }
    if (pDev->Type == 2) {
      //
      // FT232 devices
      //
      if (pDev->State == 0) {
        //
        // Initialize device
        //
        h232 = USBH_FT232_Open(pDev->DevIndex);
        if (h232 == 0) {
          pDev->Type = 0;
          continue;
        }
        //
        // Retrieve the information about the  device
        // and print them out.
        //
        USBH_FT232_GetDeviceInfo(h232, &DeviceInfo232);
        USBH_Logf_Application("Vendor  Id = 0x%0.4X", DeviceInfo232.VendorId);
        USBH_Logf_Application("Product Id = 0x%0.4X", DeviceInfo232.ProductId);
        USBH_Logf_Application("Speed      = %s", _GetPortSpeed(DeviceInfo232.Speed));
        //
        // Known device ?
        //
        for (i = 0; i < SEGGER_COUNTOF(_KnownDevices); i++) {
          if (_KnownDevices[i].VID == DeviceInfo232.VendorId &&
              _KnownDevices[i].PID == DeviceInfo232.ProductId) {
            break;
          }
        }
        if (i >= SEGGER_COUNTOF(_KnownDevices)) {
          //
          // Not found
          //
          USBH_Logf_Application("Unknown device, ignored");
          pDev->State = -1;
          continue;
        }
        pDev->CtrlMask =  _KnownDevices[i].CtrlMask;
        pDev->u.Handle232 = h232;
        //
        // Configure the FT232 device.
        //
        USBH_FT232_SetTimeouts(h232, 500, 500);
        USBH_FT232_AllowShortRead(h232, 1);
        USBH_FT232_SetDataCharacteristics(h232, USBH_FT232_BITS_8, USBH_FT232_STOP_BITS_1, USBH_FT232_PARITY_NONE);
        USBH_FT232_SetBaudRate(h232, USBH_FT232_BAUD_9600);
        USBH_FT232_SetFlowControl(h232, USBH_FT232_FLOW_XON_XOFF, 0x11, 0x13);
        USBH_FT232_SetRts(h232);
        USBH_FT232_SetDtr(h232);
        pDev->State = 50;
        pDev->Dirty = 1;
        continue;
      }
      if (pDev->State < 0) {
        USBH_FT232_Close(pDev->u.Handle232);
        pDev->Type = 0;
        continue;
      }
      if (pDev->Dirty != 0 || Dirty != 0) {
        //
        // Update display
        //
        pDev->Dirty = 0;
        for (i = 0; i < USBH_TERM_LINES; i++) {
          if (_SendCtrlSeq(pDev, _PosSequence, SEGGER_COUNTOF(_PosSequence), i) != 0) {
            break;
          }
          _Shift(i, Buff);
          USBH_FT232_Write(pDev->u.Handle232, Buff, USBH_TERM_COLUMS, NULL);
        }
      }
    }
  }
  OS_TerminateTask(NULL);
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/*********************************************************************
*
*       USBH_TERM_Print
*
*  Function description
*    Print text to terminal buffer.
*
*  Parameters
*    Line       : Line on the display to print on (Counting from 0).
*    XOffset    : X Offset of the print position.
*    sText      : 0-terminated string to print.
*/
void USBH_TERM_Print(unsigned Line, int XOffset, const char *sText) {
  unsigned Len;
  unsigned Off;

  if (XOffset >= USBH_TERM_VIRT_COLUMS || Line >= USBH_TERM_LINES) {
    return;
  }
  if (XOffset < 0) {
    Off = 0;
  } else {
    Off = XOffset;
  }
  Len = strlen(sText);
  if (Off + Len > USBH_TERM_VIRT_COLUMS) {
    Len = USBH_TERM_VIRT_COLUMS - Off;
  }
  OS_Use(&_Mutex);
  if (XOffset < 0) {
    USBH_MEMSET(_aScreen[Line], ' ', USBH_TERM_VIRT_COLUMS);
    _aLen[Line] = 0;
    _aOff[Line] = 0;
    if (Len < USBH_TERM_COLUMS) {
      Off = (USBH_TERM_COLUMS - Len) / 2;
    }
  }
  USBH_MEMCPY(&_aScreen[Line][Off], sText, Len);
  Len += Off;
  if (XOffset == 0) {
    _aLen[Line] = Len;
  } else {
    if (Len > _aLen[Line]) {
      _aLen[Line] = Len;
    }
  }
  _Touch();
  OS_Unuse(&_Mutex);
}

/*********************************************************************
*
*       USBH_TERM_Scroll
*
*  Function description
*    Scroll display contents.
*/
void USBH_TERM_Scroll(void) {
  unsigned i;

  OS_Use(&_Mutex);
  for (i = 0; i < USBH_TERM_LINES - 1; i++) {
    USBH_MEMCPY(_aScreen[i], _aScreen[i+1], USBH_TERM_VIRT_COLUMS);
    _aLen[i] = _aLen[i + 1];
  }
  USBH_MEMSET(_aScreen[USBH_TERM_LINES - 1], ' ', USBH_TERM_VIRT_COLUMS);
  _aLen[USBH_TERM_LINES - 1] = 0;
  _Touch();
  OS_Unuse(&_Mutex);
}

/*********************************************************************
*
*       USBH_TERM_Init
*
*  Function description
*    Initialize terminal functions.
*/
void USBH_TERM_Init(void) {
  USBH_MEMSET(_aScreen, ' ', sizeof(_aScreen));
  USBH_MEMSET(_DevList, 0, sizeof(_DevList));
  USBH_CDC_Init();
  //
  // The following function call makes sure that even "broken" devices or host controllers are working.
  // Some devices return invalid information in the CDC descriptors regarding the control and data interface.
  // In order to allow those device we disable the interface check and allow any devices.
  // (-> set the USBH_CDC_DISABLE_INTERFACE_CHECK flag)
  // Furthermore some USB host controller are unable to process multiple read/write transactions
  // simultaneously (eg. KinetisFS host controller).
  // For that case we disable the INTERRUPT EP transaction.
  // Be aware that asynchronous operations on such host controllers will not work properly when processing more than one.
  //
  USBH_CDC_SetConfigFlags(USBH_CDC_IGNORE_INT_EP | USBH_CDC_DISABLE_INTERFACE_CHECK);
  USBH_CDC_AddNotification(&CDCNotiHook, _cbOnAddRemoveDevice, &_DevCDC);
  USBH_FT232_Init();
  USBH_FT232_AddNotification(&FT232NotiHook, _cbOnAddRemoveDevice, &_DevFT232);
  USBH_FT232_AddCustomDeviceMask(_aVendorIds, _aProductIds, SEGGER_COUNTOF(_aVendorIds));
  _Run = 1;
  OS_CREATERSEMA(&_Mutex);
  OS_CREATETASK(&_TCBTerm, "USBH_Term", _TERM_Task, OS_GetPriority(NULL), _StackTerm);    // Start USBH TERM task
}

/*********************************************************************
*
*       USBH_TERM_Exit
*
*  Function description
*    Unregisters and de-initializes the TERM API module from emUSB-Host.
*/
void USBH_TERM_Exit(void) {
  _Run = 0;
  OS_TerminateTask(&_TCBTerm);
  USBH_CDC_RemoveNotification(&CDCNotiHook);
  USBH_CDC_Exit();
  USBH_FT232_RemoveNotification(&FT232NotiHook);
  USBH_FT232_Exit();
  OS_DeleteRSema(&_Mutex);
}

/*************************** End of file ****************************/
